# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
from os.path import dirname, join

from apprise import (
    Apprise,
    AppriseAsset,
    AppriseAttachment,
    AppriseConfig,
    NotificationManager,
    common,
)
from apprise.decorators import notify
from apprise.decorators.base import CustomNotifyPlugin

logging.disable(logging.CRITICAL)

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()

TEST_VAR_DIR = join(dirname(__file__), "var")


def test_notify_simple_decoration():
    """decorators: Test simple @notify"""

    # Verify our schema we're about to declare doesn't already exist
    # in our schema map:
    assert "utiltest" not in N_MGR

    verify_obj = {}

    # Define a function here on the spot
    @notify(on="utiltest", name="Apprise @notify Decorator Testing")
    def my_inline_notify_wrapper(
        body, title, notify_type, attach, *args, **kwargs
    ):

        # Test our body (always present)
        assert isinstance(body, str)

        # Ensure content is of type utf-8
        assert isinstance(body.encode("utf-8"), bytes)

        if attach:
            # attachment is always of type AppriseAttach
            assert isinstance(attach, AppriseAttachment)

        # Populate our object we can use to validate
        verify_obj.update(
            {
                "body": body,
                "title": title,
                "notify_type": notify_type,
                "attach": attach,
                "args": args,
                "kwargs": kwargs,
            }
        )

    # Now after our hook being inline... it's been loaded
    assert "utiltest" in N_MGR

    # Create ourselves an apprise object
    aobj = Apprise()

    assert aobj.add("utiltest://") is True

    assert len(verify_obj) == 0

    assert (
        aobj.notify(
            "Hello World",
            title="My Title",
            # add some attachments too
            attach=(
                join(TEST_VAR_DIR, "apprise-test.gif"),
                join(TEST_VAR_DIR, "apprise-test.png"),
            ),
        )
        is True
    )

    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == "Hello World"
    assert verify_obj["title"] == "My Title"
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert isinstance(verify_obj["attach"], AppriseAttachment)
    assert len(verify_obj["attach"]) == 2

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(verify_obj["kwargs"], dict)
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"]["meta"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 4
    assert "tag" in verify_obj["kwargs"]["meta"]

    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)

    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    assert verify_obj["kwargs"]["meta"]["url"] == "utiltest://"

    # Reset our verify object (so it can be populated again)
    verify_obj = {}

    # Send unicode
    assert aobj.notify("ツ".encode()) is True
    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == "ツ"  # content comes back as str (utf-8)
    assert verify_obj["title"] == ""
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert verify_obj["attach"] is None

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(verify_obj["kwargs"], dict)
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"]["meta"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 4
    assert "tag" in verify_obj["kwargs"]["meta"]

    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)

    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    assert verify_obj["kwargs"]["meta"]["url"] == "utiltest://"

    # Reset our verify object (so it can be populated again)
    verify_obj = {}

    # Send utf-8 string
    assert aobj.notify("ツ") is True

    assert len(verify_obj) > 0
    assert verify_obj["body"] == "ツ"  # content comes back as str (utf-8)
    assert verify_obj["title"] == ""
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert verify_obj["attach"] is None

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(verify_obj["kwargs"], dict)
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"]["meta"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 4
    assert "tag" in verify_obj["kwargs"]["meta"]

    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)

    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    assert verify_obj["kwargs"]["meta"]["url"] == "utiltest://"

    # Some cases that will fail internal validation:
    # - No Body
    assert aobj.notify("") is False
    # - Title only
    assert aobj.notify("", title="hello world!") is False

    # Reset our verify object (so it can be populated again)
    verify_obj = {}

    # No Body but has attachment (valid)
    assert (
        aobj.notify("", attach=(join(TEST_VAR_DIR, "apprise-test.png"),))
        is True
    )

    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == ""
    assert verify_obj["title"] == ""
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert isinstance(verify_obj["attach"], AppriseAttachment)
    assert len(verify_obj["attach"]) == 1

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(verify_obj["kwargs"], dict)
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"]["meta"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 4
    assert "tag" in verify_obj["kwargs"]["meta"]

    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)

    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    assert verify_obj["kwargs"]["meta"]["url"] == "utiltest://"

    # Reset our verify object (so it can be populated again)
    verify_obj = {}

    # We'll do another test now
    assert (
        aobj.notify(
            "Hello Another World",
            title="My Other Title",
            body_format=common.NotifyFormat.HTML,
            notify_type=common.NotifyType.WARNING,
        )
        is True
    )

    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == "Hello Another World"
    assert verify_obj["title"] == "My Other Title"
    assert verify_obj["notify_type"] == common.NotifyType.WARNING
    # We have no attachments
    assert verify_obj["attach"] is None

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] == common.NotifyFormat.HTML

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 4
    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)
    assert "tag" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["tag"], set)
    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    assert verify_obj["kwargs"]["meta"]["url"] == "utiltest://"

    assert "notexc" not in N_MGR

    # Define a function here on the spot
    @notify(on="notexc", name="Apprise @notify Exception Handling")
    def my_exception_inline_notify_wrapper(
        body, title, notify_type, attach, *args, **kwargs
    ):
        raise ValueError("An exception was thrown!")

    assert "notexc" in N_MGR

    # Create ourselves an apprise object
    aobj = Apprise()

    assert aobj.add("notexc://") is True

    # Isn't handled
    assert aobj.notify("Exceptions will be thrown!") is False

    # Tidy
    N_MGR.remove("utiltest", "notexc")


def test_notify_complex_decoration():
    """decorators: Test complex @notify"""

    # Verify our schema we're about to declare doesn't already exist
    # in our schema map:
    assert "utiltest" not in N_MGR

    verify_obj = {}

    # Define a function here on the spot
    @notify(
        on="utiltest://user@myhost:23?key=value&NOT=CaseSensitive",
        name="Apprise @notify Decorator Testing",
    )
    def my_inline_notify_wrapper(
        body, title, notify_type, attach, *args, **kwargs
    ):

        # Populate our object we can use to validate
        verify_obj.update(
            {
                "body": body,
                "title": title,
                "notify_type": notify_type,
                "attach": attach,
                "args": args,
                "kwargs": kwargs,
            }
        )

    # Now after our hook being inline... it's been loaded
    assert "utiltest" in N_MGR

    # Create ourselves an apprise object
    aobj = Apprise()

    assert aobj.add("utiltest://") is True

    assert len(verify_obj) == 0

    assert (
        aobj.notify(
            "Hello World",
            title="My Title",
            # add some attachments too
            attach=(
                join(TEST_VAR_DIR, "apprise-test.gif"),
                join(TEST_VAR_DIR, "apprise-test.png"),
            ),
        )
        is True
    )

    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == "Hello World"
    assert verify_obj["title"] == "My Title"
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert isinstance(verify_obj["attach"], AppriseAttachment)
    assert len(verify_obj["attach"]) == 2

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(verify_obj["kwargs"], dict)
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"]["meta"], dict)

    assert "asset" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["asset"], AppriseAsset)

    assert "tag" in verify_obj["kwargs"]["meta"]
    assert isinstance(verify_obj["kwargs"]["meta"]["tag"], set)

    assert len(verify_obj["kwargs"]["meta"]) == 8
    # We carry all of our default arguments from the @notify's initialization
    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"

    # Case sensitivity is lost on key assignment and always made lowercase
    # however value case sensitivity is preseved.
    # this is the assembled URL based on the combined values of the default
    # parameters with values provided in the URL (user's configuration)
    assert verify_obj["kwargs"]["meta"]["url"].startswith(
        "utiltest://user@myhost:23?"
    )

    # We don't know where they get placed, so just search for their match
    assert "key=value" in verify_obj["kwargs"]["meta"]["url"]
    assert "not=CaseSensitive" in verify_obj["kwargs"]["meta"]["url"]

    # Reset our verify object (so it can be populated again)
    verify_obj = {}

    # We'll do another test now
    aobj = Apprise()

    assert aobj.add("utiltest://customhost?key=new&key2=another") is True

    assert len(verify_obj) == 0

    # Send our notification
    assert aobj.notify("Hello World", title="My Title") is True

    # Our content was populated after the notify() call
    assert len(verify_obj) > 0
    assert verify_obj["body"] == "Hello World"
    assert verify_obj["title"] == "My Title"
    assert verify_obj["notify_type"] == common.NotifyType.INFO
    assert verify_obj["attach"] is None

    # No format was defined
    assert "body_format" in verify_obj["kwargs"]
    assert verify_obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert "meta" in verify_obj["kwargs"]
    assert isinstance(verify_obj["kwargs"], dict)
    assert len(verify_obj["kwargs"]["meta"]) == 8

    # We carry all of our default arguments from the @notify's initialization
    assert verify_obj["kwargs"]["meta"]["schema"] == "utiltest"
    # Our host get's correctly over-ridden
    assert verify_obj["kwargs"]["meta"]["host"] == "customhost"

    assert verify_obj["kwargs"]["meta"]["user"] == "user"
    assert verify_obj["kwargs"]["meta"]["port"] == 23
    assert isinstance(verify_obj["kwargs"]["meta"]["qsd"], dict)
    assert len(verify_obj["kwargs"]["meta"]["qsd"]) == 3
    # our key is over-ridden
    assert verify_obj["kwargs"]["meta"]["qsd"]["key"] == "new"
    # Our other keys are preserved
    assert verify_obj["kwargs"]["meta"]["qsd"]["not"] == "CaseSensitive"
    # New keys are added
    assert verify_obj["kwargs"]["meta"]["qsd"]["key2"] == "another"

    # Case sensitivity is lost on key assignment and always made lowercase
    # however value case sensitivity is preseved.
    # this is the assembled URL based on the combined values of the default
    # parameters with values provided in the URL (user's configuration)
    assert verify_obj["kwargs"]["meta"]["url"].startswith(
        "utiltest://user@customhost:23?"
    )

    # We don't know where they get placed, so just search for their match
    assert "key=new" in verify_obj["kwargs"]["meta"]["url"]
    assert "not=CaseSensitive" in verify_obj["kwargs"]["meta"]["url"]
    assert "key2=another" in verify_obj["kwargs"]["meta"]["url"]

    # Tidy
    N_MGR.remove("utiltest")


def test_notify_decorator_urls_with_space():
    """decorators: URLs containing spaces"""
    # This is in relation to https://github.com/caronc/apprise/issues/1264

    # Verify our schema we're about to declare doesn't already exist
    # in our schema map:
    assert "post" not in N_MGR

    verify_obj = []

    @notify(on="posts")
    def apprise_custom_api_call_wrapper(
        body, title, notify_type, attach, meta, *args, **kwargs
    ):

        # Track what is added
        verify_obj.append(
            {
                "body": body,
                "title": title,
                "notify_type": notify_type,
                "attach": attach,
                "meta": meta,
                "args": args,
                "kwargs": kwargs,
            }
        )

    assert "posts" in N_MGR

    # Create ourselves an apprise object
    aobj = Apprise()

    # Add our configuration
    aobj.add("posts://example.com/my endpoint?-token=ab cdefg")

    # We loaded 1 item
    assert len(aobj) == 1

    # Nothing stored yet in our object
    assert len(verify_obj) == 0

    # Send utf-8 characters
    assert aobj.notify("ツ".encode(), title="My Title") is True

    # Service notified
    assert len(verify_obj) == 1

    # Extract our object
    obj = verify_obj.pop()

    assert obj.get("body") == "ツ"
    assert obj.get("title") == "My Title"
    assert obj.get("notify_type") == "info"
    assert obj.get("attach") is None
    assert isinstance(obj.get("args"), tuple)
    assert len(obj.get("args")) == 0
    assert obj.get("kwargs") == {"body_format": None}
    meta = obj.get("meta")
    assert isinstance(meta, dict)

    assert meta.get("schema") == "posts"
    assert (
        meta.get("url")
        == "posts://example.com/my%20endpoint?-token=ab%20cdefg"
    )
    assert meta.get("qsd") == {"-token": "ab cdefg"}
    assert meta.get("host") == "example.com"
    assert meta.get("fullpath") == "/my%20endpoint"
    assert meta.get("path") == "/"
    assert meta.get("query") == "my%20endpoint"
    assert isinstance(meta.get("tag"), set)
    assert len(meta.get("tag")) == 0
    assert isinstance(meta.get("asset"), AppriseAsset)

    # Tidy
    N_MGR.remove("posts")


def test_notify_multi_instance_decoration(tmpdir):
    """decorators: Test multi-instance @notify"""

    # Verify our schema we're about to declare doesn't already exist
    # in our schema map:
    assert "multi" not in N_MGR

    verify_obj = []

    # Define a function here on the spot
    @notify(on="multi", name="Apprise @notify Decorator Testing")
    def my_inline_notify_wrapper(
        body, title, notify_type, attach, meta, *args, **kwargs
    ):

        assert isinstance(body, str)

        # Track what is added
        verify_obj.append(
            {
                "body": body,
                "title": title,
                "notify_type": notify_type,
                "attach": attach,
                "meta": meta,
                "args": args,
                "kwargs": kwargs,
            }
        )

    # Now after our hook being inline... it's been loaded
    assert "multi" in N_MGR

    # Prepare our config
    t = tmpdir.mkdir("multi-test").join("apprise.yml")
    t.write("""urls:
    - multi://user1:pass@hostname
    - multi://user2:pass2@hostname?verify=no
    """)

    # Create ourselves a config object
    ac = AppriseConfig(paths=str(t))

    # Create ourselves an apprise object
    aobj = Apprise()

    # Add our configuration
    aobj.add(ac)

    # The number of configuration files that exist
    assert len(ac) == 1

    # 2 notification endpoints are loaded
    assert len(ac.servers()) == 2

    # Nothing stored yet in our object
    assert len(verify_obj) == 0

    # Send utf-8 characters
    assert aobj.notify("ツ".encode(), title="My Title") is True

    assert len(verify_obj) == 2

    # Python 3.6 does not nessisarily return list in order
    # So let's be sure it's sorted by the user id field to make the remaining
    # checks on this test easy
    verify_obj = sorted(verify_obj, key=lambda x: x["meta"]["user"])

    # Our content was populated after the notify() call
    obj = verify_obj[0]
    assert obj["body"] == "ツ"
    assert obj["title"] == "My Title"
    assert obj["notify_type"] == common.NotifyType.INFO

    meta = obj["meta"]
    assert isinstance(meta, dict)

    # No format was defined
    assert "body_format" in obj["kwargs"]
    assert obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(obj["kwargs"], dict)

    assert "asset" in meta
    assert isinstance(meta["asset"], AppriseAsset)

    assert "tag" in meta
    assert isinstance(meta["tag"], set)

    assert len(meta) == 8
    # We carry all of our default arguments from the @notify's initialization
    assert meta["schema"] == "multi"
    assert meta["host"] == "hostname"
    assert meta["user"] == "user1"
    assert meta["verify"] is True
    assert meta["password"] == "pass"

    # Verify our URL is correct
    assert meta["url"] == "multi://user1:pass@hostname"

    #
    # Now verify our second URL saved correct
    #

    # Our content was populated after the notify() call
    obj = verify_obj[1]
    assert obj["body"] == "ツ"
    assert obj["title"] == "My Title"
    assert obj["notify_type"] == common.NotifyType.INFO

    meta = obj["meta"]
    assert isinstance(meta, dict)

    # No format was defined
    assert "body_format" in obj["kwargs"]
    assert obj["kwargs"]["body_format"] is None

    # The meta argument allows us to further parse the URL parameters
    # specified
    assert isinstance(obj["kwargs"], dict)

    assert "asset" in meta
    assert isinstance(meta["asset"], AppriseAsset)

    assert "tag" in meta
    assert isinstance(meta["tag"], set)

    assert len(meta) == 9
    # We carry all of our default arguments from the @notify's initialization
    assert meta["schema"] == "multi"
    assert meta["host"] == "hostname"
    assert meta["user"] == "user2"
    assert meta["password"] == "pass2"
    assert meta["verify"] is False
    assert meta["qsd"]["verify"] == "no"

    # Verify our URL is correct
    assert meta["url"] == "multi://user2:pass2@hostname?verify=no"

    # Tidy
    N_MGR.remove("multi")


def test_custom_notify_plugin_decoration():
    """decorators: CustomNotifyPlugin testing"""

    CustomNotifyPlugin()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise.plugins.clickatell import NotifyClickatell

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "clickatell://",
        {
            # only schema provided
            "instance": TypeError,
        },
    ),
    (
        "clickatell:///",
        {
            # invalid apikey
            "instance": TypeError,
        },
    ),
    (
        "clickatell://@/",
        {
            # invalid apikey
            "instance": TypeError,
        },
    ),
    (
        "clickatell://{}@/".format("1" * 10),
        {
            # no api key provided
            "instance": TypeError,
        },
    ),
    (
        "clickatell://{}@{}/".format("1" * 3, "a" * 32),
        {
            # invalid From/Source
            "instance": TypeError
        },
    ),
    (
        "clickatell://{}/".format("a" * 32),
        {
            # no targets provided
            "instance": NotifyClickatell,
            # We have no one to notify
            "notify_response": False,
        },
    ),
    (
        "clickatell://{}@{}/".format("1" * 10, "a" * 32),
        {
            # no targets provided (no one to notify)
            "instance": NotifyClickatell,
            # We have no one to notify
            "notify_response": False,
        },
    ),
    (
        "clickatell://{}@{}/123/{}/abcd".format("1" * 10, "a" * 32, "3" * 15),
        {
            # valid everything but target numbers
            "instance": NotifyClickatell,
            # We have no one to notify
            "notify_response": False,
        },
    ),
    (
        "clickatell://{}/{}".format("1" * 10, "a" * 32),
        {
            # everything valid (no source defined)
            "instance": NotifyClickatell,
            # We have no one to notify
            "notify_response": False,
        },
    ),
    (
        "clickatell://{}@{}/{}".format("1" * 10, "a" * 32, "1" * 10),
        {
            # everything valid
            "instance": NotifyClickatell,
        },
    ),
    (
        "clickatell://{}/{}".format("a" * 32, "1" * 10),
        {
            # everything valid (no source)
            "instance": NotifyClickatell,
        },
    ),
    (
        "clickatell://_?apikey={}&from={}&to={},{}".format(
            "a" * 32, "1" * 10, "1" * 10, "1" * 10
        ),
        {
            # use get args to accomplish the same thing
            "instance": NotifyClickatell,
        },
    ),
    (
        "clickatell://_?apikey={}".format("a" * 32),
        {
            # use get args
            "instance": NotifyClickatell,
            "notify_response": False,
        },
    ),
    (
        "clickatell://_?apikey={}&from={}".format("a" * 32, "1" * 10),
        {
            # use get args
            "instance": NotifyClickatell,
            "notify_response": False,
        },
    ),
    (
        "clickatell://{}@{}/{}".format("1" * 10, "a" * 32, "1" * 10),
        {
            "instance": NotifyClickatell,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "clickatell://{}@{}/{}".format("1" * 10, "a" * 32, "1" * 10),
        {
            "instance": NotifyClickatell,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_clickatell_urls():
    """NotifyClickatell() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_clickatell_edge_cases(mock_post):
    """NotifyClickatell() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # Initialize some generic (but valid) apikeys
    apikey = "b" * 32
    from_phone = "+1 (555) 123-3456"

    # No apikey specified
    with pytest.raises(TypeError):
        NotifyClickatell(apikey=None, from_phone=from_phone)

    # a error response
    response.status_code = 400
    response.content = dumps(
        {
            "code": 21211,
            "message": "The 'To' number +1234567 is not a valid phone number.",
        }
    )
    mock_post.return_value = response

    # Initialize our object
    obj = NotifyClickatell(apikey=apikey, from_phone=from_phone)

    # We will fail with the above error code
    assert obj.notify("title", "body", "info") is False



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import loads
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise
from apprise.plugins.evolution import NotifyEvolution

logging.disable(logging.CRITICAL)

# ---------------------------------------------------------------------------
# URL-based tests — exercised by AppriseURLTester
# ---------------------------------------------------------------------------

apprise_url_tests = (
    # No host — parse_url returns None
    (
        "evolution://",
        {
            "instance": None,
        },
    ),
    # No host, explicit empty credentials
    (
        "evolution://:@/",
        {
            "instance": None,
        },
    ),
    # Missing apikey (no user field)
    (
        "evolution://hostname/myinstance/5511999999999",
        {
            "instance": TypeError,
        },
    ),
    # Missing instance (single path entry treated as instance; no phone left)
    (
        "evolution://myapikey@hostname/5511999999999",
        {
            "instance": TypeError,
        },
    ),
    # Missing phone number
    (
        "evolution://myapikey@hostname/myinstance",
        {
            "instance": TypeError,
        },
    ),
    # Invalid phone number
    (
        "evolution://myapikey@hostname/myinstance/notaphone",
        {
            "instance": TypeError,
        },
    ),
    # Valid HTTP — minimal
    (
        "evolution://myapikey@hostname/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "privacy_url": (
                "evolution://m...y@hostname/myinstance/5511999999999"
            ),
        },
    ),
    # Valid HTTPS
    (
        "evolutions://myapikey@hostname/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "privacy_url": (
                "evolutions://m...y@hostname/myinstance/5511999999999"
            ),
        },
    ),
    # Custom port
    (
        "evolution://myapikey@hostname:8080/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "privacy_url": (
                "evolution://m...y@hostname:8080/myinstance/5511999999999"
            ),
        },
    ),
    # Multiple targets
    (
        "evolution://myapikey@hostname/myinstance/5511999999999/5521888888888",
        {
            "instance": NotifyEvolution,
        },
    ),
    # Target via ?to=
    (
        "evolution://myapikey@hostname/myinstance/5511999999999"
        "?to=5521888888888",
        {
            "instance": NotifyEvolution,
        },
    ),
    # Force HTTP error
    (
        "evolution://myapikey@hostname/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    # Unknown HTTP response code
    (
        "evolution://myapikey@hostname/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    # Connection exceptions
    (
        "evolution://myapikey@hostname/myinstance/5511999999999",
        {
            "instance": NotifyEvolution,
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_evolution_urls():
    """NotifyEvolution() Apprise URLs."""
    AppriseURLTester(tests=apprise_url_tests).run_all()


# ---------------------------------------------------------------------------
# Edge cases
# ---------------------------------------------------------------------------


def test_plugin_evolution_parse_url_no_path():
    """NotifyEvolution.parse_url() returns instance=None when path is empty."""
    result = NotifyEvolution.parse_url("evolution://key@host")
    assert result is not None
    assert result["instance"] is None


def test_plugin_evolution_edge_cases():
    """NotifyEvolution() direct instantiation edge cases."""

    # No apikey
    with pytest.raises(TypeError):
        NotifyEvolution(
            apikey=None, instance="inst", targets=["5511999999999"]
        )

    # Blank apikey
    with pytest.raises(TypeError):
        NotifyEvolution(
            apikey="   ", instance="inst", targets=["5511999999999"]
        )

    # No instance
    with pytest.raises(TypeError):
        NotifyEvolution(apikey="key", instance=None, targets=["5511999999999"])

    # No targets
    with pytest.raises(TypeError):
        NotifyEvolution(apikey="key", instance="inst", targets=[])

    # All targets invalid
    with pytest.raises(TypeError):
        NotifyEvolution(
            apikey="key", instance="inst", targets=["notaphone", "xx"]
        )


# ---------------------------------------------------------------------------
# send() payload and header verification
# ---------------------------------------------------------------------------


@mock.patch("requests.post")
def test_plugin_evolution_send(mock_post):
    """NotifyEvolution() verifies POST payload and headers."""

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    obj = Apprise.instantiate(
        "evolution://MYAPIKEY@localhost:8080/myinstance/5511999999999"
    )
    assert isinstance(obj, NotifyEvolution)

    assert obj.notify(body="Hello World", title="Test") is True
    assert mock_post.call_count == 1

    # Verify URL
    call = mock_post.call_args
    assert "localhost:8080" in call[0][0]
    assert "/message/sendText/myinstance" in call[0][0]

    # Verify headers
    headers = call[1]["headers"]
    assert headers["apikey"] == "MYAPIKEY"
    assert headers["Content-Type"] == "application/json"

    # Verify payload
    payload = loads(call[1]["data"])
    assert payload["number"] == "5511999999999"
    assert "Hello World" in payload["text"]


@mock.patch("requests.post")
def test_plugin_evolution_multiple_targets(mock_post):
    """NotifyEvolution() sends one request per target."""

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    obj = Apprise.instantiate(
        "evolution://key@host/inst/5511111111111/5522222222222/5533333333333"
    )
    assert isinstance(obj, NotifyEvolution)
    assert len(obj) == 3

    assert obj.notify(body="msg") is True
    assert mock_post.call_count == 3

    numbers = [loads(c[1]["data"])["number"] for c in mock_post.call_args_list]
    assert "5511111111111" in numbers
    assert "5522222222222" in numbers
    assert "5533333333333" in numbers


@mock.patch("requests.post")
def test_plugin_evolution_partial_failure(mock_post):
    """NotifyEvolution() returns False if any target fails."""

    r_ok = requests.Request()
    r_ok.status_code = requests.codes.ok
    r_ok.content = b""

    r_err = requests.Request()
    r_err.status_code = requests.codes.internal_server_error
    r_err.content = b"error"

    mock_post.side_effect = [r_ok, r_err]

    obj = Apprise.instantiate(
        "evolution://key@host/inst/5511111111111/5522222222222"
    )
    assert obj.notify(body="msg") is False
    assert mock_post.call_count == 2


@mock.patch("requests.post")
def test_plugin_evolution_title_in_body(mock_post):
    """NotifyEvolution() merges the title into the body via Apprise."""

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    obj = Apprise.instantiate("evolution://key@host/inst/5511999999999")
    assert obj.notify(body="body text", title="My Title") is True

    payload = loads(mock_post.call_args[1]["data"])
    assert "My Title" in payload["text"]
    assert "body text" in payload["text"]


@mock.patch("requests.post")
def test_plugin_evolution_https(mock_post):
    """NotifyEvolution() uses HTTPS when evolutions:// schema is given."""

    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    obj = Apprise.instantiate(
        "evolutions://key@secure.host/inst/5511999999999"
    )
    assert isinstance(obj, NotifyEvolution)
    assert obj.secure is True

    assert obj.notify(body="secure msg") is True
    url = mock_post.call_args[0][0]
    assert url.startswith("https://")



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
import sys

from apprise import emojis

logging.disable(logging.CRITICAL)

# Ensure we don't create .pyc files for these tests
sys.dont_write_bytecode = True


def test_emojis():
    "emojis: apply_emojis() testing"

    assert emojis.apply_emojis("") == ""
    assert emojis.apply_emojis("no change") == "no change"
    assert emojis.apply_emojis(":smile:") == "😄"
    assert emojis.apply_emojis(":smile::smile:") == "😄😄"

    # Missing Delimiters
    assert emojis.apply_emojis(":smile") == ":smile"
    assert emojis.apply_emojis("smile:") == "smile:"

    # Bad data
    assert emojis.apply_emojis(None) == ""
    assert emojis.apply_emojis(object) == ""
    assert emojis.apply_emojis(True) == ""
    assert emojis.apply_emojis(4.0) == ""


def test_emojis_alternation_aliases():
    """emojis: apply_emojis() handles regex alternation keys correctly."""

    # thumbsup / +1
    assert emojis.apply_emojis(":thumbsup:") == "👍"
    assert emojis.apply_emojis(":+1:") == "👍"

    # thumbsdown / -1
    assert emojis.apply_emojis(":thumbsdown:") == "👎"
    assert emojis.apply_emojis(":-1:") == "👎"

    # laughing / satisfied
    assert emojis.apply_emojis(":laughing:") == "😆"
    assert emojis.apply_emojis(":satisfied:") == "😆"

    # poop aliases
    assert emojis.apply_emojis(":poop:") == "💩"
    assert emojis.apply_emojis(":hankey:") == "💩"
    assert emojis.apply_emojis(":shit:") == "💩"

    # boom / collision
    assert emojis.apply_emojis(":boom:") == "💥"
    assert emojis.apply_emojis(":collision:") == "💥"

    # exclamation aliases
    assert emojis.apply_emojis(":exclamation:") == "❗"
    assert emojis.apply_emojis(":heavy_exclamation_mark:") == "❗"

    # knife / hocho
    assert emojis.apply_emojis(":knife:") == "🔪"
    assert emojis.apply_emojis(":hocho:") == "🔪"

    # memo / pencil
    assert emojis.apply_emojis(":memo:") == "📝"
    assert emojis.apply_emojis(":pencil:") == "📝"

    # uk / gb
    assert emojis.apply_emojis(":uk:") == "🇬🇧"
    assert emojis.apply_emojis(":gb:") == "🇬🇧"

    # aliases embedded in text
    assert emojis.apply_emojis("hello :thumbsup: world") == "hello 👍 world"
    assert emojis.apply_emojis(":poop::boom:") == "💩💥"

    # optional-group patterns: ':hand:' and ':raised_hand:' both map to ✋
    assert emojis.apply_emojis(":hand:") == "✋"
    assert emojis.apply_emojis(":raised_hand:") == "✋"

    # runner / running
    assert emojis.apply_emojis(":runner:") == "🏃"
    assert emojis.apply_emojis(":running:") == "🏃"


def test_emojis_lookup_no_pattern_match():
    """emojis: _lookup() returns the raw token when _EMOJI_PATTERN_LIST is
    empty (defensive fallback -- return text branch)."""

    # Ensure the engine is initialised first.
    emojis.apply_emojis(":smile:")

    original = emojis._EMOJI_PATTERN_LIST
    emojis._EMOJI_PATTERN_LIST = []
    try:
        # With no patterns to match against, _lookup falls through to
        # "return text", so the matched token is passed back unchanged.
        result = emojis.apply_emojis(":smile:")
        assert result == ":smile:"
    finally:
        emojis._EMOJI_PATTERN_LIST = original



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys
from unittest import mock

import pytest

import apprise
from apprise import NotifyBase
from apprise.plugins.growl import GrowlPriority, NotifyGrowl

try:
    from gntp import errors

    TEST_GROWL_EXCEPTIONS = (
        errors.NetworkError(0, "gntp.ParseError() not handled"),
        errors.AuthError(0, "gntp.AuthError() not handled"),
        errors.ParseError(0, "gntp.ParseError() not handled"),
        errors.UnsupportedError(0, "gntp.UnsupportedError() not handled"),
    )

except ImportError:
    # no problem; gntp isn't available to us
    pass

# Disable logging for a cleaner testing output
import logging

logging.disable(logging.CRITICAL)


@pytest.mark.skipif(
    "gntp" in sys.modules, reason="Requires that gntp NOT be installed"
)
def test_plugin_growl_gntp_import_error():
    """NotifyGrowl() Import Error."""
    # If the object is disabled, then it can't be instantiated
    obj = apprise.Apprise.instantiate("growl://growl.server")
    assert obj is None


@pytest.mark.skipif("gntp" not in sys.modules, reason="Requires gntp")
@mock.patch("gntp.notifier.GrowlNotifier")
def test_plugin_growl_exception_handling(mock_gntp):
    """NotifyGrowl() Exception Handling."""
    TEST_GROWL_EXCEPTIONS = (
        errors.NetworkError(0, "gntp.ParseError() not handled"),
        errors.AuthError(0, "gntp.AuthError() not handled"),
        errors.ParseError(0, "gntp.ParseError() not handled"),
        errors.UnsupportedError(0, "gntp.UnsupportedError() not handled"),
    )

    mock_notifier = mock.Mock()
    mock_gntp.return_value = mock_notifier
    mock_notifier.notify.return_value = True

    # First we test the growl.register() function
    for exception in TEST_GROWL_EXCEPTIONS:
        mock_notifier.register.side_effect = exception

        # instantiate our object
        obj = apprise.Apprise.instantiate(
            "growl://growl.server.hostname", suppress_exceptions=False
        )

        # Verify Growl object was instantiated
        assert obj is not None

        # We will fail to send the notification because our registration
        # would have failed
        assert (
            obj.notify(
                title="test", body="body", notify_type=apprise.NotifyType.INFO
            )
            is False
        )

    # Now we test the growl.notify() function
    mock_notifier.register.side_effect = None
    for exception in TEST_GROWL_EXCEPTIONS:
        mock_notifier.notify.side_effect = exception

        # instantiate our object
        obj = apprise.Apprise.instantiate(
            "growl://growl.server.hostname", suppress_exceptions=False
        )

        # Verify Growl object was instantiated
        assert obj is not None

        # We will fail to send the notification because of the underlining
        # notify() call throws an exception
        assert (
            obj.notify(
                title="test", body="body", notify_type=apprise.NotifyType.INFO
            )
            is False
        )


@pytest.mark.skipif("gntp" not in sys.modules, reason="Requires gntp")
@mock.patch("gntp.notifier.GrowlNotifier")
def test_plugin_growl_general(mock_gntp):
    """NotifyGrowl() General Checks."""

    urls = (
        ##################################
        # NotifyGrowl
        ##################################
        (
            "growl://",
            {
                "instance": None,
            },
        ),
        ("growl://:@/", {"instance": None}),
        (
            "growl://pass@growl.server",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://ignored:pass@growl.server",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.server",
            {
                "instance": NotifyGrowl,
                # don't include an image by default
                "include_image": False,
            },
        ),
        (
            "growl://growl.server?version=1",
            {
                "instance": NotifyGrowl,
            },
        ),
        # Test sticky flag
        (
            "growl://growl.server?sticky=yes",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.server?sticky=no",
            {
                "instance": NotifyGrowl,
            },
        ),
        # Force a failure
        (
            "growl://growl.server?version=1",
            {
                "instance": NotifyGrowl,
                "growl_response": None,
            },
        ),
        (
            "growl://growl.server?version=2",
            {
                # don't include an image by default
                "include_image": False,
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.server?version=2",
            {
                # don't include an image by default
                "include_image": False,
                "instance": NotifyGrowl,
                "growl_response": None,
            },
        ),
        # Priorities
        (
            "growl://pass@growl.server?priority=low",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://pass@growl.server?priority=moderate",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://pass@growl.server?priority=normal",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://pass@growl.server?priority=high",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://pass@growl.server?priority=emergency",
            {
                "instance": NotifyGrowl,
            },
        ),
        # Invalid Priorities
        (
            "growl://pass@growl.server?priority=invalid",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://pass@growl.server?priority=",
            {
                "instance": NotifyGrowl,
            },
        ),
        # invalid version
        (
            "growl://growl.server?version=",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.server?version=crap",
            {
                "instance": NotifyGrowl,
            },
        ),
        # Ports
        (
            "growl://growl.changeport:2000",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.garbageport:garbage",
            {
                "instance": NotifyGrowl,
            },
        ),
        (
            "growl://growl.colon:",
            {
                "instance": NotifyGrowl,
            },
        ),
    )

    # iterate over our dictionary and test it out
    for url, meta in urls:
        # Our expected instance
        instance = meta.get("instance", None)

        # Our expected exception
        exception = meta.get("exception", None)

        # Our expected server objects
        self = meta.get("self", None)

        # Our expected Query response (True, False, or exception type)
        response = meta.get("response", True)

        # Allow us to force the server response code to be something other then
        # the defaults
        growl_response = meta.get("growl_response", bool(response))

        mock_notifier = mock.Mock()
        mock_gntp.return_value = mock_notifier
        mock_notifier.notify.side_effect = None

        # Store our response
        mock_notifier.notify.return_value = growl_response

        try:
            obj = apprise.Apprise.instantiate(url, suppress_exceptions=False)

            assert exception is None

            if obj is None:
                # We're done
                continue

            if instance is None:
                # Expected None but didn't get it
                raise AssertionError()

            assert isinstance(obj, instance) is True

            # Test our URL Identifier is generated
            assert isinstance(obj.url_id(), str) is True

            if isinstance(obj, NotifyBase):
                # We loaded okay; now lets make sure we can reverse this url
                assert isinstance(obj.url(), str) is True

                # Test our privacy=True flag
                assert isinstance(obj.url(privacy=True), str) is True

                # Instantiate the exact same object again using the URL from
                # the one that was already created properly
                obj_cmp = apprise.Apprise.instantiate(obj.url())

                # Our object should be the same instance as what we had
                # originally expected above.
                if not isinstance(obj_cmp, NotifyBase):
                    # Assert messages are hard to trace back with the way
                    # these tests work. Just printing before throwing our
                    # assertion failure makes things easier to debug later on
                    raise AssertionError()

            if self:
                # Iterate over our expected entries inside of our object
                for key, val in self.items():
                    # Test that our object has the desired key
                    assert hasattr(key, obj)
                    assert getattr(key, obj) == val

            try:
                # check that we're as expected
                assert (
                    obj.notify(
                        title="test",
                        body="body",
                        notify_type=apprise.NotifyType.INFO,
                    )
                    == response
                )

            except Exception as e:
                # Check that we were expecting this exception to happen
                assert isinstance(e, response)

        except AssertionError:
            # Don't mess with these entries
            raise

        except Exception as e:
            # Handle our exception
            assert exception is not None
            assert isinstance(e, exception)


@pytest.mark.skipif("gntp" not in sys.modules, reason="Requires gntp")
@mock.patch("gntp.notifier.GrowlNotifier")
def test_plugin_growl_config_files(mock_gntp):
    """NotifyGrowl() Config File Cases."""
    content = """
    urls:
      - growl://pass@growl.server:
          - priority: -2
            tag: growl_int low
          - priority: "-2"
            tag: growl_str_int low
          - priority: low
            tag: growl_str low

          # This will take on moderate (default) priority
          - priority: invalid
            tag: growl_invalid

      - growl://pass@growl.server:
          - priority: 2
            tag: growl_int emerg
          - priority: "2"
            tag: growl_str_int emerg
          - priority: emergency
            tag: growl_str emerg
    """

    mock_notifier = mock.Mock()
    mock_gntp.return_value = mock_notifier
    mock_notifier.notify.return_value = True

    # Create ourselves a config object
    ac = apprise.AppriseConfig()
    assert ac.add_config(content=content) is True

    aobj = apprise.Apprise()

    # Add our configuration
    aobj.add(ac)

    # We should be able to read our 7 servers from that
    # 3x low
    # 3x emerg
    # 1x invalid (so takes on normal priority)
    assert len(ac.servers()) == 7
    assert len(aobj) == 7
    assert len(list(aobj.find(tag="low"))) == 3
    for s in aobj.find(tag="low"):
        assert s.priority == GrowlPriority.LOW

    assert len(list(aobj.find(tag="emerg"))) == 3
    for s in aobj.find(tag="emerg"):
        assert s.priority == GrowlPriority.EMERGENCY

    assert len(list(aobj.find(tag="growl_str"))) == 2
    assert len(list(aobj.find(tag="growl_str_int"))) == 2
    assert len(list(aobj.find(tag="growl_int"))) == 2

    assert len(list(aobj.find(tag="growl_invalid"))) == 1
    assert (
        next(aobj.find(tag="growl_invalid")).priority == GrowlPriority.NORMAL
    )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import sys
import types
from unittest.mock import Mock, call

from helpers import reload_plugin
import pytest

import apprise
from apprise.plugins.glib import GLibUrgency, NotifyGLib

# Disable logging output during testing
logging.disable(logging.CRITICAL)


@pytest.fixture
def enabled_glib_environment(monkeypatch):
    """
    Fully mocked GI/GLib/Gio/GdkPixbuf environment for local and CI.
    """
    # Step 1: Fake gi and repository
    gi = types.ModuleType("gi")
    gi.require_version = Mock()

    fake_variant = Mock(name="Variant")
    fake_error = type("GLibError", (Exception,), {})
    fake_pixbuf = Mock()
    fake_image = Mock()

    fake_pixbuf.new_from_file.return_value = fake_image
    fake_image.get_width.return_value = 100
    fake_image.get_height.return_value = 100
    fake_image.get_rowstride.return_value = 1
    fake_image.get_has_alpha.return_value = False
    fake_image.get_bits_per_sample.return_value = 8
    fake_image.get_n_channels.return_value = 1
    fake_image.get_pixels.return_value = b""

    gi.repository = types.SimpleNamespace(
        Gio=Mock(),
        GLib=types.SimpleNamespace(Variant=fake_variant, Error=fake_error),
        GdkPixbuf=types.SimpleNamespace(Pixbuf=fake_pixbuf),
    )

    # Step 2: Inject into sys.modules
    sys.modules["gi"] = gi
    sys.modules["gi.repository"] = gi.repository

    # Step 3: Reload plugin with all mocks in place
    reload_plugin("glib")


def test_plugin_glib_gdkpixbuf_attribute_error(monkeypatch):
    """Simulate AttributeError from importing GdkPixbuf"""

    # Create gi module
    gi = types.ModuleType("gi")

    # Create gi.repository mock, but DO NOT include GdkPixbuf
    gi.repository = types.SimpleNamespace(
        Gio=Mock(),
        GLib=types.SimpleNamespace(
            Variant=Mock(), Error=type("GLibError", (Exception,), {})
        ),
        # GdkPixbuf missing entirely triggers AttributeError
    )

    def fake_require_version(name, version):
        if name == "GdkPixbuf":
            # Simulate success in require_version
            return
        return

    gi.require_version = Mock(side_effect=fake_require_version)

    # Inject into sys.modules
    sys.modules["gi"] = gi
    sys.modules["gi.repository"] = gi.repository

    # Trigger the plugin reload with our patched environment
    reload_plugin("glib")

    from apprise.plugins import glib as plugin_glib

    assert plugin_glib.NOTIFY_GLIB_IMAGE_SUPPORT is False


def test_plugin_glib_basic_notify(enabled_glib_environment):
    """Basic notification path"""
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    assert isinstance(obj, NotifyGLib)
    assert obj.notify("body", title="title") is True


def test_plugin_glib_url_includes_coordinates(enabled_glib_environment):
    """Test that x/y coordinates appear in the rendered URL."""
    obj = apprise.Apprise.instantiate(
        "glib://_/?x=7&y=9", suppress_exceptions=False
    )
    url = obj.url(privacy=False)

    assert "x=7" in url
    assert "y=9" in url


def test_plugin_glib_icon_fails_gracefully(mocker, enabled_glib_environment):
    """Simulate image load failure"""
    import gi

    gi.repository.GdkPixbuf.Pixbuf.new_from_file.side_effect = AttributeError(
        "fail"
    )
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    spy = mocker.spy(obj, "logger")
    assert obj.notify("msg", title="t") is True
    assert any(
        "Could not load notification icon" in str(x)
        for x in spy.warning.call_args_list
    )


def test_plugin_glib_send_raises_glib_error(mocker, enabled_glib_environment):
    """Simulate GLib.Error in DBusProxy creation"""
    import gi

    gi.repository.Gio.DBusProxy.new_for_bus_sync.side_effect = (
        gi.repository.GLib.Error("fail")
    )
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    assert obj.notify("fail test") is False


def test_plugin_glib_send_raises_generic(mocker, enabled_glib_environment):
    """Simulate generic error in gio_iface.Notify()"""
    # Re: https://github.com/caronc/apprise/issues/1383
    # This test validates that the NotifyGLib plugin correctly handles a
    # generic exception raised by the `Notify()` method call on a mocked
    # DBus interface. However, it is only meaningful in environments that:
    #
    #  1. Do NOT have PyGObject (`gi`) installed, OR
    #  2. Have `gi`, but without introspection or live bindings activated.
    #
    # When PyGObject is installed and active, the `gi.repository` namespace
    # becomes populated by introspected C-based objects that do not behave
    # like regular Python functions. This causes mock patching via
    # `mocker.patch("gi.repository.Gio.DBusProxy.new_for_bus_sync")` to
    # silently fail or be ignored, as Python's mocking machinery cannot
    # reliably override these introspected symbols.
    #
    # This test exists to ensure coverage of legacy or minimal environments
    # where Apprise's GLib support can still be used through soft mocks,
    # such as CI/CD pipelines or headless test setups where PyGObject is
    # absent or stubbed (as done via `enabled_glib_environment`).
    #
    # Note: In production environments with active PyGObject, exception
    # handling is already tested via `GLib.Error` branches or during actual
    # usage of `NotifyGLib.send()`. This test supplements that by simulating
    # the rare fallback case of a non-GLib-related exception during Notify().
    import gi

    if hasattr(gi, "repository"):
        pytest.skip(
            "pygobject introspection active, test won't behave as expected"
        )

    fake_iface = Mock()
    fake_iface.Notify.side_effect = RuntimeError("boom")

    mocker.patch(
        "gi.repository.Gio.DBusProxy.new_for_bus_sync",
        return_value=fake_iface,
    )

    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    logger = mocker.spy(obj, "logger")
    assert obj.notify("boom", title="fail") is False
    logger.warning.assert_called_with("Failed to send GLib/Gio notification.")


def test_plugin_glib_disabled(mocker, enabled_glib_environment):
    """Test disabled plugin returns False on notify()"""
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    obj.enabled = False
    assert obj.notify("x") is False


def test_plugin_glib_invalid_coords():
    """Invalid x/y coordinates cause TypeError"""
    with pytest.raises(TypeError):
        NotifyGLib(x_axis="bad", y_axis="1")
    with pytest.raises(TypeError):
        NotifyGLib(x_axis="1", y_axis="bad")


def test_plugin_glib_urgency_parsing():
    """Urgency variants map correctly"""
    assert NotifyGLib(urgency="high").urgency == GLibUrgency.HIGH
    assert NotifyGLib(urgency="invalid").urgency == GLibUrgency.NORMAL
    assert NotifyGLib(urgency="2").urgency == GLibUrgency.HIGH
    assert NotifyGLib(urgency=0).urgency == GLibUrgency.LOW


def test_plugin_glib_parse_url_fields():
    url = "glib://_/?x=5&y=5&image=no&priority=high"
    result = NotifyGLib.parse_url(url)
    assert result["x_axis"] == "5"
    assert result["y_axis"] == "5"
    assert result["include_image"] is False
    assert result["urgency"] == "high"


def test_plugin_glib_xy_axis_applied_to_variant(enabled_glib_environment):
    """Ensure x/y values are added to GLib.Variant payload."""
    obj = apprise.Apprise.instantiate(
        "glib://_/?x=5&y=10", suppress_exceptions=False
    )

    # Patch GLib.Variant to track calls
    import gi

    spy_variant = Mock(wraps=gi.repository.GLib.Variant)
    gi.repository.GLib.Variant = spy_variant

    assert obj.notify("Test with coords", title="xy") is True

    # Check x and y were added to meta_payload
    assert call("i", 5) in spy_variant.mock_calls
    assert call("i", 10) in spy_variant.mock_calls


def test_plugin_glib_no_image_support(monkeypatch, enabled_glib_environment):
    """Simulate GdkPixbuf unavailable"""
    monkeypatch.setattr(
        "apprise.plugins.glib.NOTIFY_GLIB_IMAGE_SUPPORT", False
    )
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    assert obj.notify("no image") is True


def test_plugin_glib_url_redaction(enabled_glib_environment):
    """url() privacy mode redacts safely"""
    obj = apprise.Apprise.instantiate(
        "glib://_/?image=no&urgency=high", suppress_exceptions=False
    )
    url = obj.url(privacy=True)
    assert "image=" in url
    assert "urgency=" in url
    assert url.startswith("glib://_/")


def test_plugin_glib_require_version_importerror(monkeypatch):
    """Simulate gi.require_version() raising ImportError"""
    gi = types.ModuleType("gi")
    gi.require_version = Mock(side_effect=ImportError("no gio"))
    sys.modules["gi"] = gi
    reload_plugin("glib")
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    assert not isinstance(obj, NotifyGLib)


def test_plugin_glib_require_version_valueerror(monkeypatch):
    """Simulate gi.require_version() raising ValueError without reload
    crash."""

    import gi

    import apprise.plugins.glib as plugin_glib

    # Patch require_version after import
    monkeypatch.setattr(
        gi, "require_version", Mock(side_effect=ValueError("fail"))
    )

    # Re-evaluate plugin support logic manually
    try:
        gi.require_version("Gio", "2.0")

    except Exception:
        plugin_glib.NOTIFY_GLIB_SUPPORT_ENABLED = False
        plugin_glib.NotifyGLib.enabled = False

    # Confirm plugin is now marked disabled
    assert not plugin_glib.NotifyGLib.enabled

    # Apprise will skip this plugin
    obj = apprise.Apprise.instantiate("glib://", suppress_exceptions=False)
    assert not isinstance(obj, plugin_glib.NotifyGLib)


def test_plugin_glib_gdkpixbuf_require_version_valueerror(monkeypatch):
    """Simulate gi.require_version('GdkPixbuf', ...) raising ValueError"""

    # Step 1: Mock GI
    gi = types.ModuleType("gi")
    gi.repository = types.ModuleType("gi.repository")

    def fake_require_version(name: str, version: str) -> None:
        if name == "GdkPixbuf":
            raise ValueError("GdkPixbuf unavailable")

    gi.require_version = Mock(side_effect=fake_require_version)

    # Step 2: Patch into sys.modules
    sys.modules["gi"] = gi
    sys.modules["gi.repository"] = gi.repository

    # Step 3: Reload plugin to trigger branch
    reload_plugin("glib")

    # Step 4: Confirm GdkPixbuf image support was not enabled
    from apprise.plugins import glib as plugin_glib

    assert plugin_glib.NOTIFY_GLIB_IMAGE_SUPPORT is False


def test_plugin_glib_notify_generic_exception(
    mocker, enabled_glib_environment
):
    """
    Test that a generic exception occurring during the notification send
    is caught and handled (returning False).
    """
    import gi

    # 1. Create a mock object to act as the DBus Proxy
    mock_proxy = mocker.Mock()

    # 2. Force the Notify method to raise a generic Exception
    #    This triggers the 'except Exception as e:'
    mock_proxy.Notify.side_effect = Exception("Simulated Generic Failure")

    # 3. Patch the Gio constructor to return our crashing mock
    #    The code calls: Gio.DBusProxy.new_for_bus_sync(...)
    gi.repository.Gio.DBusProxy.new_for_bus_sync.return_value = mock_proxy

    # 4. Instantiate the plugin
    from apprise.plugins.glib import NotifyGLib

    obj = NotifyGLib(targets=["glib://"])

    # 5. Spy on the logger to ensure the warning is logged
    logger_spy = mocker.spy(obj, "logger")

    # 6. Execute notify()
    #    It should catch the exception and return False
    assert obj.notify(title="Title", body="Body") is False

    # 7. Verify the specific log message from glib.py
    logger_spy.warning.assert_called_with(
        "Failed to send GLib/Gio notification."
    )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from inspect import cleandoc

# Disable logging for a cleaner testing output
import logging
import re
import sys
import threading
import types

import pytest

from apprise import Apprise, NotificationManager
from apprise.plugins import NotifyBase

logging.disable(logging.CRITICAL)

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()


def test_notification_manager_general():
    """
    N_MGR: Notification Manager General testing

    """
    # Clear our set so we can test init calls
    N_MGR.unload_modules()
    assert isinstance(N_MGR.schemas(), list)
    assert len(N_MGR.schemas()) > 0
    N_MGR.unload_modules(disable_native=True)
    assert isinstance(N_MGR.schemas(), list)
    assert len(N_MGR.schemas()) == 0

    N_MGR.unload_modules()
    assert len(N_MGR) > 0

    N_MGR.unload_modules()
    iter(N_MGR)
    iter(N_MGR)

    N_MGR.unload_modules()
    assert bool(N_MGR) is False
    assert len(list(iter(N_MGR))) > 0
    assert bool(N_MGR)

    N_MGR.unload_modules()
    assert isinstance(N_MGR.plugins(), types.GeneratorType)
    assert len(list(N_MGR.plugins())) > 0
    N_MGR.unload_modules(disable_native=True)
    assert isinstance(N_MGR.plugins(), types.GeneratorType)
    assert len(list(N_MGR.plugins())) == 0
    N_MGR.unload_modules()
    assert isinstance(N_MGR["json"](host="localhost"), NotifyBase)
    N_MGR.unload_modules()
    assert "json" in N_MGR

    # Define our good:// url
    class DisabledNotification(NotifyBase):
        # Always disabled
        enabled = False

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def notify(self, *args, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support url() function
            return ""

    # Define our good:// url
    class GoodNotification(NotifyBase):
        secure_protocol = ("good", "goods")

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def notify(self, *args, **kwargs):
            # Pretend everything is okay
            return True

        def url(self, **kwargs):
            # Support url() function
            return ""

    N_MGR.unload_modules()
    assert N_MGR.add(GoodNotification)
    assert "good" in N_MGR
    assert "goods" in N_MGR
    assert "abcd" not in N_MGR
    assert "xyz" not in N_MGR

    N_MGR.unload_modules()
    assert N_MGR.add(GoodNotification, "abcd")
    assert "good" in N_MGR
    assert "goods" in N_MGR
    assert "abcd" in N_MGR
    assert "xyz" not in N_MGR

    N_MGR.unload_modules()
    assert N_MGR.add(GoodNotification, ["abcd", "xYz"])
    assert "good" in N_MGR
    assert "goods" in N_MGR
    assert "abcd" in N_MGR
    # Lower case
    assert "xyz" in N_MGR

    N_MGR.unload_modules()
    # Not going to work; schemas must be a list of string
    assert N_MGR.add(GoodNotification, object) is False

    N_MGR.unload_modules()
    with pytest.raises(KeyError):
        del N_MGR["good"]
    N_MGR["good"] = GoodNotification
    del N_MGR["good"]

    N_MGR.unload_modules()
    N_MGR["good"] = GoodNotification
    assert N_MGR["good"].enabled is True
    N_MGR.enable_only("json", "xml")
    assert N_MGR["good"].enabled is False
    assert N_MGR["json"].enabled is True
    assert N_MGR["jsons"].enabled is True
    assert N_MGR["xml"].enabled is True
    assert N_MGR["xmls"].enabled is True

    # Only two plugins are enabled
    assert len(list(N_MGR.plugins(include_disabled=False))) == 2

    N_MGR.enable_only("good")
    assert N_MGR["good"].enabled is True
    assert N_MGR["json"].enabled is False
    assert N_MGR["jsons"].enabled is False
    assert N_MGR["xml"].enabled is False
    assert N_MGR["xmls"].enabled is False

    assert len(list(N_MGR.plugins(include_disabled=False))) == 1

    N_MGR.unload_modules()
    N_MGR["disabled"] = DisabledNotification
    assert N_MGR["disabled"].enabled is False
    N_MGR.enable_only("disabled")
    # Can't enable items that aren't supposed to be:
    assert N_MGR["disabled"].enabled is False

    N_MGR["good"] = GoodNotification
    assert N_MGR["good"].enabled is True

    # You can't disable someething already disabled
    N_MGR.disable("disabled")
    assert N_MGR["disabled"].enabled is False

    N_MGR.unload_modules()
    N_MGR.enable_only("form", "xml")
    for schema in N_MGR.schemas(include_disabled=False):
        assert re.match(r"^(form|xml)s?$", schema, re.IGNORECASE) is not None

    N_MGR.unload_modules()
    assert N_MGR["form"].enabled is True
    assert N_MGR["xml"].enabled is True
    assert N_MGR["json"].enabled is True
    N_MGR.enable_only("form", "xml")
    assert N_MGR["form"].enabled is True
    assert N_MGR["xml"].enabled is True
    assert N_MGR["json"].enabled is False

    N_MGR.disable("invalid", "xml")
    assert N_MGR["form"].enabled is True
    assert N_MGR["xml"].enabled is False
    assert N_MGR["json"].enabled is False

    # Detect that our json object is enabled
    with pytest.raises(KeyError):
        # The below can not be indexed
        N_MGR["invalid"]

    N_MGR.unload_modules()
    N_MGR.disable("invalid", "xml")

    N_MGR.unload_modules()
    assert N_MGR["json"].enabled is True

    # Work with an empty module tree
    N_MGR.unload_modules(disable_native=True)
    with pytest.raises(KeyError):
        # The below can not be indexed
        N_MGR["good"]

    N_MGR.unload_modules()
    assert "hello" not in N_MGR
    assert "good" not in N_MGR
    assert "goods" not in N_MGR

    N_MGR["hello"] = GoodNotification
    assert "hello" in N_MGR
    assert "good" in N_MGR
    assert "goods" in N_MGR

    N_MGR.unload_modules()
    N_MGR["good"] = GoodNotification

    with pytest.raises(KeyError):
        # Can not assign the value again without getting a Conflict
        N_MGR["good"] = GoodNotification

    N_MGR.unload_modules()
    N_MGR.remove("good", "invalid")
    assert "good" not in N_MGR
    assert "goods" not in N_MGR


def test_notification_manager_add_force_overrides_schema_without_unload():
    """Verify add(force=True) overrides existing schema without unloading."""
    import sys

    from apprise.plugins import N_MGR

    class NotifyDiscordCustom:
        """A minimal custom discord override used for testing."""

        protocol = "discord"
        secure_protocol = None
        service_name = "Discord (Custom)"

        def __init__(self, *args, **kwargs):
            pass

        def send(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return "discord://"

    # Ensure native modules are loaded
    N_MGR.unload_modules()
    N_MGR.load_modules()

    # Confirm 'discord' is available and capture its module name
    assert "discord" in N_MGR
    native_plugin = N_MGR["discord"]
    native_module = native_plugin.__module__
    assert native_module in sys.modules

    # A normal add should fail due to the conflict
    assert N_MGR.add(NotifyDiscordCustom, schemas="discord") is False

    # A forced add should succeed and must not unload the native module
    assert (
        N_MGR.add(NotifyDiscordCustom, schemas="discord", force=True) is True
    )
    assert N_MGR["discord"] is NotifyDiscordCustom
    assert native_module in sys.modules


def test_notification_manager_module_loading(tmpdir):
    """
    N_MGR: Notification Manager Module Loading

    """

    # Handle loading modules twice (they gracefully handle not loading more in
    # memory then needed)
    N_MGR.load_modules()
    N_MGR.load_modules()

    #
    # Thread Testing
    #

    # This tests against a racing condition when the modules have not been
    # loaded.  When multiple instances of Apprise are all instantiated,
    # the loading of the modules will occur for each instance if detected
    # having not been previously done, this tests that we can dynamically
    # support the loading of modules once whe multiple instances to apprise
    # are instantiated.
    thread_count = 10

    def thread_test(result, no):
        """Load our apprise object with valid URLs and store our result."""
        apobj = Apprise()
        result[no] = (
            apobj.add("json://localhost")
            and apobj.add("form://localhost")
            and apobj.add("xml://localhost")
        )

    # Unload our modules
    N_MGR.unload_modules()

    # Prepare threads to load
    results = [None] * thread_count
    threads = [
        threading.Thread(target=thread_test, args=(results, no))
        for no in range(thread_count)
    ]

    # Verify we can safely load our modules in a thread safe environment
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    # Verify we loaded our urls in all threads successfully
    for result in results:
        assert result is True


def test_notification_manager_decorators(tmpdir):
    """
    N_MGR: Notification Manager Decorator testing

    """

    # Prepare ourselves a file to work with
    notify_hook = tmpdir.mkdir("goodmodule").join("__init__.py")
    notify_hook.write(
        cleandoc("""
    from apprise.decorators import notify

    # We want to trigger on anyone who configures a call to clihook://
    @notify(on="clihooka")
    def mywrapper(body, title, notify_type, *args, **kwargs):
        # A simple test - print to screen
        print("A {}: {} - {}".format(notify_type, title, body))

        # No return (so a return of None) get's translated to True

    # Define another in the same file; uppercase goes to lower
    @notify(on="CLIhookb")
    def mywrapper(body, title, notify_type, *args, **kwargs):
        # A simple test - print to screen
        print("B {}: {} - {}".format(notify_type, title, body))

        # No return (so a return of None) get's translated to True
    """)
    )

    N_MGR.module_detection(str(notify_hook))

    assert "clihooka" in N_MGR
    assert "clihookb" in N_MGR
    N_MGR.unload_modules()
    assert "clihooka" not in N_MGR
    assert "clihookb" not in N_MGR

    N_MGR.module_detection(str(notify_hook))
    assert "clihooka" in N_MGR
    assert "clihookb" in N_MGR
    del N_MGR["clihookb"]
    assert "clihooka" in N_MGR
    assert "clihookb" not in N_MGR
    del N_MGR["clihooka"]
    assert "clihooka" not in N_MGR
    assert "clihookb" not in N_MGR

    # Prepare ourselves a file to work with
    notify_base = tmpdir.mkdir("plugins")
    notify_test = notify_base.join("NotifyTest.py")
    notify_test.write(
        cleandoc("""
    #
    # Bare Minimum Valid Object
    #
    from apprise.plugins import NotifyBase
    from apprise.common import NotifyType

    class NotifyTest(NotifyBase):

        service_name = 'Test'

        # The services URL
        service_url = 'https://github.com/caronc/apprise/'

        # Define our protocol
        secure_protocol = 'myservice'

        # A URL that takes you to the setup/help of the specific protocol
        setup_url = 'https://appriseit.com/services/myservice/'

        # Define object templates
        templates = (
            '{schema}://',
        )

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

        def send(self, body, title='', notify_type=NotifyType.INFO, **kwargs):
            return True

        def url(self):
            return 'myservice://'
    """)
    )
    assert "myservice" not in N_MGR
    N_MGR.load_modules(path=str(notify_base))
    assert "myservice" in N_MGR
    del N_MGR["myservice"]
    assert "myservice" not in N_MGR

    assert "myservice" not in N_MGR
    N_MGR.load_modules(path=str(notify_base))

    # It's still not loaded because the path has already been scanned
    assert "myservice" not in N_MGR
    N_MGR.load_modules(path=str(notify_base), force=True)
    assert "myservice" in N_MGR

    # Double load will test section of code that prevents a notification
    # From reloading if previously already loaded
    N_MGR.load_modules(path=str(notify_base))
    # Our item is still loaded as expected
    assert "myservice" in N_MGR

    # Simple test to make sure we can handle duplicate entries loaded
    N_MGR.load_modules(path=str(notify_base), force=True)
    N_MGR.load_modules(path=str(notify_base), force=True)


def test_notification_manager_add_force_returns_false_if_conflict_persists(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """
    N_MGR.add(force=True) must still fail safely if conflicts persist after the
    attempted unmap.

    This explicitly targets the defensive re-check branch in add().
    """
    # Ensure native modules are loaded and we have a known schema to collide on
    N_MGR.unload_modules()
    N_MGR.load_modules()
    assert "discord" in N_MGR

    class NotifyDiscordCustom:
        protocol = "discord"
        secure_protocol = None
        service_name = "Discord (Custom)"

        def __init__(self, *args, **kwargs) -> None:
            pass

        def send(self, *args, **kwargs) -> bool:
            return True

        def url(self, **kwargs) -> str:
            return "discord://"

    # Simulate an unmap failure by making remove() a no-op.
    # This ensures the conflict remains on the re-check and triggers the
    # warning + return False branch.
    def _noop_remove(*args, **kwargs) -> None:
        return None

    monkeypatch.setattr(N_MGR, "remove", _noop_remove)

    assert (
        N_MGR.add(NotifyDiscordCustom, schemas="discord", force=True) is False
    )


# ---------------------------------------------------------------------------
# Helpers shared by the runtime_deps / eviction tests
# ---------------------------------------------------------------------------


def _make_dep_plugin(schema, *deps):
    """Return a minimal NotifyBase subclass that declares *deps in
    runtime_deps() and is always enabled."""

    class _Plugin(NotifyBase):
        protocol = schema
        enabled = True

        def notify(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return f"{schema}://"

        @staticmethod
        def runtime_deps():
            return deps

    _Plugin.__name__ = f"Notify_{schema}"
    return _Plugin


# ---------------------------------------------------------------------------


def test_manager_dep_counter_built_on_load():
    """_dep_counter is populated after native modules are loaded."""
    N_MGR.unload_modules()
    # Trigger native load
    assert len(N_MGR.schemas()) > 0

    # The counter must exist and be a dict
    assert isinstance(N_MGR._dep_counter, dict)

    # Every library declared by an enabled plugin must appear in the counter
    for plugin in N_MGR.plugins(include_disabled=False):
        for lib in plugin.runtime_deps():
            assert lib in N_MGR._dep_counter
            assert N_MGR._dep_counter[lib] >= 1


def test_manager_dep_counter_custom_plugins():
    """_dep_counter reflects only enabled plugins with runtime_deps."""
    N_MGR.unload_modules()

    PlugA = _make_dep_plugin("plugA", "fakelib1")
    PlugB = _make_dep_plugin("plugB", "fakelib1", "fakelib2")
    PlugC = _make_dep_plugin("plugC")  # no deps

    N_MGR["plugA"] = PlugA
    N_MGR["plugB"] = PlugB
    N_MGR["plugC"] = PlugC

    # Rebuild counter to include newly registered plugins
    N_MGR._build_dep_counter()

    assert N_MGR._dep_counter.get("fakelib1", 0) >= 2
    assert N_MGR._dep_counter.get("fakelib2", 0) >= 1
    # PlugC has no deps so it contributes nothing
    assert "fakelib_none" not in N_MGR._dep_counter


def test_manager_disable_decrements_counter():
    """disable() decrements the dep counter for the disabled plugin's libs."""
    N_MGR.unload_modules()

    PlugA = _make_dep_plugin("solodep", "uniquelib99")
    N_MGR["solodep"] = PlugA
    N_MGR._build_dep_counter()

    assert N_MGR._dep_counter.get("uniquelib99", 0) >= 1
    before = N_MGR._dep_counter["uniquelib99"]

    N_MGR.disable("solodep")
    assert N_MGR._dep_counter.get("uniquelib99", 0) == before - 1


def test_manager_shared_dep_not_evicted_until_all_disabled():
    """A shared library is not evicted until every plugin using it is off."""
    import sys

    N_MGR.unload_modules()
    N_MGR.evict_on_disable = True

    # Two plugins share the same fake library
    PlugX = _make_dep_plugin("sharedX", "sharedfakelib")
    PlugY = _make_dep_plugin("sharedY", "sharedfakelib")

    # Inject a fake module so eviction has something to remove
    fake_mod = types.ModuleType("sharedfakelib")
    sys.modules["sharedfakelib"] = fake_mod

    N_MGR["sharedX"] = PlugX
    N_MGR["sharedY"] = PlugY
    N_MGR._build_dep_counter()

    assert N_MGR._dep_counter.get("sharedfakelib", 0) >= 2

    # Disable one — library should still be present (counter > 0)
    N_MGR.disable("sharedX")
    assert "sharedfakelib" in sys.modules

    # Disable the other — counter should hit 0 and library gets evicted
    N_MGR.disable("sharedY")
    assert "sharedfakelib" not in sys.modules

    N_MGR.evict_on_disable = False


def test_manager_evict_on_disable_false_by_default():
    """evict_on_disable defaults to False; no eviction even at counter 0."""
    import sys

    N_MGR.unload_modules()
    assert N_MGR.evict_on_disable is False

    PlugZ = _make_dep_plugin("nodiscard", "nodiscardlib")
    fake_mod = types.ModuleType("nodiscardlib")
    sys.modules["nodiscardlib"] = fake_mod

    N_MGR["nodiscard"] = PlugZ
    N_MGR._build_dep_counter()

    N_MGR.disable("nodiscard")

    # Library is still present because evict_on_disable is False
    assert "nodiscardlib" in sys.modules

    # Cleanup
    del sys.modules["nodiscardlib"]


def test_manager_evict_library_graceful_keyerror():
    """_evict_library handles a missing sys.modules key gracefully."""
    import sys

    N_MGR.unload_modules()

    # Put a module in, then remove it manually before eviction to simulate
    # a race or double-eviction scenario
    fake_mod = types.ModuleType("ghostlib")
    sys.modules["ghostlib"] = fake_mod
    del sys.modules["ghostlib"]

    # Should not raise even though the key is gone
    N_MGR._evict_library("ghostlib")


def test_manager_evict_library_removes_submodules():
    """_evict_library removes the top-level package and all its submodules."""
    import sys

    N_MGR.unload_modules()

    root = types.ModuleType("mypkg")
    sub = types.ModuleType("mypkg.sub")
    deep = types.ModuleType("mypkg.sub.deep")

    sys.modules["mypkg"] = root
    sys.modules["mypkg.sub"] = sub
    sys.modules["mypkg.sub.deep"] = deep

    N_MGR._evict_library("mypkg")

    assert "mypkg" not in sys.modules
    assert "mypkg.sub" not in sys.modules
    assert "mypkg.sub.deep" not in sys.modules


def test_manager_enable_only_updates_dep_counter():
    """enable_only() decrements counters for disabled plugins.

    We isolate this test to fake libraries only to avoid evicting real native
    extensions (e.g. cryptography/PyO3) that cannot be safely re-imported in
    the same interpreter session.
    """
    import sys

    N_MGR.unload_modules()
    N_MGR.evict_on_disable = True

    PlugKeep = _make_dep_plugin("keepme2", "keeplib2")
    PlugDrop = _make_dep_plugin("dropme2", "droplib2")

    fake_keep = types.ModuleType("keeplib2")
    fake_drop = types.ModuleType("droplib2")
    sys.modules["keeplib2"] = fake_keep
    sys.modules["droplib2"] = fake_drop

    N_MGR["keepme2"] = PlugKeep
    N_MGR["dropme2"] = PlugDrop

    # Seed the dep counter for only our two fake plugins so that disabling
    # dropme2 drives the droplib2 counter to zero — without touching real
    # native-library plugins loaded alongside them.
    N_MGR._dep_counter["keeplib2"] = 1
    N_MGR._dep_counter["droplib2"] = 1

    # Disable dropme2 directly; keepme2 stays enabled.
    N_MGR.disable("dropme2")

    assert N_MGR["keepme2"].enabled is True
    assert N_MGR["dropme2"].enabled is False
    assert "keeplib2" in sys.modules  # counter still 1 — not evicted
    assert "droplib2" not in sys.modules  # counter hit 0 — evicted

    # Cleanup
    del sys.modules["keeplib2"]
    N_MGR.evict_on_disable = False


def test_manager_unload_resets_dep_counter():
    """unload_modules() resets _dep_counter to an empty dict.

    Seeds the counter via a fake plugin rather than relying on native
    plugins whose optional libraries may not be installed (the minimal
    tox env disables all such plugins, leaving _dep_counter empty after
    a native load).
    """
    N_MGR.unload_modules()
    # Seed counter directly so the post-unload assertion is meaningful
    PlugR = _make_dep_plugin("resettest_xyz", "resetlib_xyz")
    N_MGR["resettest_xyz"] = PlugR
    N_MGR._build_dep_counter()
    assert N_MGR._dep_counter.get("resetlib_xyz", 0) >= 1

    N_MGR.unload_modules()
    assert N_MGR._dep_counter == {}


def test_manager_known_plugin_runtime_deps():
    """Plugins known to carry optional deps must declare runtime_deps()."""
    N_MGR.unload_modules()

    from apprise.plugins.blink1 import NotifyBlink1
    from apprise.plugins.fcm import NotifyFCM
    from apprise.plugins.growl import NotifyGrowl
    from apprise.plugins.mqtt import NotifyMQTT
    from apprise.plugins.simplepush import NotifySimplePush
    from apprise.plugins.smpp import NotifySMPP
    from apprise.plugins.vapid import NotifyVapid
    from apprise.plugins.xmpp.base import NotifyXMPP

    assert "hid" in NotifyBlink1.runtime_deps()
    assert "paho" in NotifyMQTT.runtime_deps()
    assert "gntp" in NotifyGrowl.runtime_deps()
    assert "smpplib" in NotifySMPP.runtime_deps()
    assert "slixmpp" in NotifyXMPP.runtime_deps()
    assert "cryptography" in NotifySimplePush.runtime_deps()
    assert "cryptography" in NotifyFCM.runtime_deps()
    assert "cryptography" in NotifyVapid.runtime_deps()

    # Base class has no deps
    assert NotifyBase.runtime_deps() == ()


# ===========================================================================
# Coverage: additional tests for previously uncovered branches
# ===========================================================================


def test_manager_len_already_loaded():
    """Verify __len__ works correctly when modules are already loaded."""
    N_MGR.unload_modules()
    # First call: triggers lazy load
    _ = len(N_MGR)
    # Second call: modules already loaded, returns length directly
    assert len(N_MGR) > 0


def test_manager_unload_orphan_custom_module():
    """Verify unload_modules() skips custom module map entries whose
    name is not present in the native module map."""
    N_MGR.unload_modules()
    N_MGR.load_modules()
    # Inject a synthetic entry whose 'name' value is not a key in
    # _module_map so the early-continue branch is exercised.
    N_MGR._custom_module_map["orphan_key_xyz"] = {
        "name": "orphan_module_xyz_not_native",
        "path": "/fake/path.py",
        "notify": {},
    }
    N_MGR.unload_modules()
    assert "orphan_key_xyz" not in N_MGR._custom_module_map


def test_module_detection_invalid_paths():
    """Verify module_detection returns early for invalid path inputs."""
    N_MGR.unload_modules()
    N_MGR.module_detection(None)
    N_MGR.module_detection([])
    N_MGR.module_detection(123)


def test_module_detection_nonexistent_path():
    """Verify a path that does not exist is silently skipped."""
    N_MGR.unload_modules()
    N_MGR.module_detection(["/tmp/__no_such_path_xyz_abc_123__"])


def test_module_detection_directory_scan(tmpdir):
    """Scans a plain directory (no __init__.py at root) containing several
    fixture files that each exercise a different branch."""
    N_MGR.unload_modules()

    # File whose name starts with '.' -> module_re fails -> ignored
    tmpdir.join(".hidden_xyz").write("")

    # File without .py extension -> valid_python_file_re fails inside
    # _import_module -> skipped
    tmpdir.join("noext_xyz").write("")

    # Valid .py but raises ImportError at module level ->
    # import_module returns None -> skipped
    tmpdir.join("badimport_xyz.py").write(
        "import this_library_does_not_exist_xyz_abc\n"
    )

    # Valid .py with no @notify -> module_pyname absent from
    # _custom_module_map after import -> skipped
    tmpdir.join("nohooks_xyz.py").write("x = 1\n")

    # Subdir without __init__.py -> no __init__.py -> skipped
    tmpdir.mkdir("subpkg_no_init_xyz")

    # Subdir WITH __init__.py, pre-populated in cache so the cache-hit
    # branch is exercised (file exists -> don't skip, but cache -> skip
    # _import_module call)
    cached_sub = tmpdir.mkdir("subpkg_cached_xyz")
    cached_init = cached_sub.join("__init__.py")
    cached_init.write("# cached\n")
    N_MGR._paths_previously_scanned.add(str(cached_init))

    N_MGR.module_detection(str(tmpdir))
    N_MGR.unload_modules()


def test_module_detection_package_dir(tmpdir):
    """Directory with __init__.py: already-scanned path is skipped;
    fresh path is imported but yields no hooks."""
    N_MGR.unload_modules()

    # A package dir (has __init__.py) with no @notify hooks
    pkgdir = tmpdir.mkdir("mypkg_xyz")
    pkg_path = str(pkgdir)
    init_path = str(pkgdir.join("__init__.py"))
    pkgdir.join("__init__.py").write("# no hooks\n")

    # Pre-populate the cache with the __init__.py path so the
    # already-scanned-inner-path branch is exercised on first call
    N_MGR._paths_previously_scanned.add(init_path)
    N_MGR.module_detection(pkg_path)

    # Discard both paths
    # so the add-to-cache branch is exercised on the second call
    N_MGR._paths_previously_scanned.discard(init_path)
    N_MGR._paths_previously_scanned.discard(pkg_path)
    N_MGR.module_detection(pkg_path)
    N_MGR.unload_modules()


def test_module_detection_non_matching_file(tmpdir):
    """File path whose basename fails module_re is silently ignored."""
    N_MGR.unload_modules()

    # A filename starting with '.' does not match module_re -> ignored
    hidden = tmpdir.join(".hidden_direct_xyz")
    hidden.write("")

    N_MGR.module_detection(str(hidden))
    N_MGR.unload_modules()


def test_module_detection_reload_existing(tmpdir):
    """_import_module clears and re-registers a custom module that was
    previously loaded into _custom_module_map."""
    N_MGR.unload_modules()

    hook = tmpdir.mkdir("reloadpkg_xyz").join("__init__.py")
    hook.write(
        cleandoc("""
        from apprise.decorators import notify

        @notify(on="reloadhookxyz")
        def handler(body, title, notify_type, *args, **kwargs):
            pass
    """)
    )

    # First load — populates _custom_module_map
    N_MGR.module_detection(str(hook))
    assert "reloadhookxyz" in N_MGR

    # Remove path from cache so the same file is processed a second time
    N_MGR._paths_previously_scanned.discard(str(hook))

    # Second load -> _import_module finds module_pyname already in
    # _custom_module_map -> clear + re-register
    N_MGR.module_detection(str(hook))
    assert "reloadhookxyz" in N_MGR
    N_MGR.unload_modules()


def test_build_dep_counter_empty_map():
    """_build_dep_counter returns early when _module_map is empty."""
    # disable_native=True sets _module_map = {} (falsy) -> returns early
    N_MGR.unload_modules(disable_native=True)
    N_MGR._build_dep_counter()
    assert N_MGR._dep_counter == {}


def test_build_dep_counter_no_runtime_deps_attr():
    """Plugin without a callable runtime_deps is skipped by
    _build_dep_counter."""
    N_MGR.unload_modules()
    N_MGR.load_modules()

    class MinimalPlugin:
        """Minimal plugin: enabled but has no runtime_deps attribute."""

        enabled = True

    N_MGR._module_map["minimal_test_xyz_657"] = {
        "plugin": {MinimalPlugin},
        "module": None,
        "path": "apprise.adhoc.minimal_test_xyz_657",
        "native": False,
    }
    # getattr returns None -> not callable(None) -> continue
    N_MGR._build_dep_counter()
    N_MGR.unload_modules()


def test_update_dep_counter_no_runtime_deps():
    """_update_dep_counter returns early when the plugin has no callable
    runtime_deps."""
    N_MGR.unload_modules()

    class NoDepPlugin:
        """Plugin with no runtime_deps attribute at all."""

    # getattr returns None -> not callable(None) -> return early
    N_MGR._update_dep_counter(NoDepPlugin, -1)


def test_evict_library_keyerror_race(monkeypatch):
    """KeyError during sys.modules eviction is handled gracefully.

    sys.modules is replaced with a subclass whose __delitem__ always
    raises KeyError.  The key is found but the delete fails, exercising
    the except-KeyError branch.  With evicted == 0 and to_remove
    non-empty the elif warning branch is also reached.
    """
    N_MGR.unload_modules()

    class BrokenModulesDict(dict):
        def __delitem__(self, key):
            raise KeyError(key)

    broken = BrokenModulesDict(sys.modules)
    broken["evict_race_lib_xyz"] = types.ModuleType("evict_race_lib_xyz")
    monkeypatch.setattr(sys, "modules", broken)

    # to_remove = ["evict_race_lib_xyz"]; del raises -> KeyError caught
    # evicted stays 0, to_remove non-empty -> warning branch reached
    N_MGR._evict_library("evict_race_lib_xyz")


def test_load_modules_import_fallback_fails(tmpdir):
    """__import__ raises ImportError and the path-based fallback also
    returns None because the file itself fails to import."""
    N_MGR.unload_modules()

    # A .py file whose top-level import always fails
    tmpdir.join("alwaysfail_xyz.py").write(
        "import this_library_does_not_exist_xyz_abc\n"
    )
    # load_modules uses the default prefix ("apprise.plugins") so
    # __import__("apprise.plugins.alwaysfail_xyz") raises ImportError,
    # then import_module(path, ...) also returns None -> skipped
    N_MGR.load_modules(path=str(tmpdir))
    N_MGR.unload_modules()


def test_load_modules_no_valid_class(tmpdir):
    """Module loads successfully but contains no qualifying class, so
    module_class stays None and the module is skipped."""
    N_MGR.unload_modules()

    # A module with only a docstring: dir() yields nothing that matches
    # module_filter_re (all dunders start with '_') -> empty list ->
    # module_class stays None -> skipped
    tmpdir.join("notifyempty_xyz.py").write('"""Empty module."""\n')

    N_MGR.load_modules(path=str(tmpdir))
    N_MGR.unload_modules()


def test_load_modules_schema_conflict(tmpdir):
    """A schema already registered is logged as a conflict and
    skipped."""
    N_MGR.unload_modules()
    N_MGR.load_modules()  # native load -> "json" in _schema_map

    tmpdir.join("notifyconflict_xyz.py").write(
        cleandoc("""
        from apprise.plugins.base import NotifyBase

        class NotifyConflict(NotifyBase):
            service_name = "Conflict"
            protocol = "json"
            secure_protocol = None
            service_url = "https://example.com"
            setup_url = "https://example.com"
            notify_url = ""
            templates = ('{schema}://',)

            def send(self, body, title="", **kwargs):
                return True

            def url(self, **kwargs):
                return "json://"
    """)
    )

    # __import__("apprise.plugins.notifyconflict_xyz") -> ImportError;
    # fallback loads the file; NotifyConflict.schemas() = {"json"};
    # "json" already in _schema_map -> conflict logged, skipped
    N_MGR.load_modules(path=str(tmpdir))
    N_MGR.unload_modules()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import contextlib
import ipaddress
import select
import socket
import ssl
import time
from typing import Optional, Union

from ..exception import AppriseException, AppriseInvalidData
from ..logger import logger

TimeoutType = Optional[Union[float, tuple[Optional[float], Optional[float]]]]


class AppriseSocketError(AppriseException):
    """Raised for socket or TLS related failures."""


class SocketTransport:
    """
    TCP client transport with optional TLS upgrade.

    Behaviour:
      - secure=False (default): plain TCP
      - secure=True: upgrade to TLS (immediately in connect(), or manually via
        start_tls())
      - verify=True (default): validate certificate chain and hostname using a
        certifi CA bundle
      - verify=False: accept invalid or self-signed certs

    Timeout behaviour (requests-compatible):
      - timeout=float => (connect, read) both set to float
      - timeout=(connect, read) => tuple form
      - None => no defaults (connect/read can block indefinitely)
    """

    def __init__(
        self,
        host: str,
        port: int,
        bind_addr: Optional[str] = None,
        bind_port: Optional[int] = None,
        secure: bool = False,
        verify: bool = True,
        timeout: TimeoutType = 10.0,
        retries: int = 0,
    ) -> None:
        self.host = host
        self.port = int(port)
        self.bind_addr = bind_addr
        self.bind_port = bind_port

        self.secure = bool(secure)
        self.verify = bool(verify)
        self.retries = retries

        self._connect_timeout, self._read_timeout = self._coerce_timeout(
            timeout
        )

        self._sock: Optional[socket.socket] = None
        self._rfile = None
        self._wfile = None
        self._is_tls: bool = False

        # True once we have successfully read or written data since the last
        # connect(). Used to decide whether reconnect attempts are allowed.
        self._had_io: bool = False

        self.local_addr: Optional[tuple[str, int]] = None
        self.remote_addr: Optional[tuple[str, int]] = None

    @staticmethod
    def _coerce_timeout(
        timeout: TimeoutType,
    ) -> tuple[Optional[float], Optional[float]]:
        """
        Coerce requests-style timeout into (connect_timeout, read_timeout).
        """
        if timeout is None:
            return None, None

        if isinstance(timeout, (int, float)):
            t = float(timeout)
            if t < 0:
                raise AppriseInvalidData("timeout must be >= 0")
            return t, t

        if isinstance(timeout, tuple) and len(timeout) == 2:
            connect_t, read_t = timeout
            if connect_t is not None:
                connect_t = float(connect_t)
                if connect_t < 0:
                    raise AppriseInvalidData("connect timeout must be >= 0")
            if read_t is not None:
                read_t = float(read_t)
                if read_t < 0:
                    raise AppriseInvalidData("read timeout must be >= 0")
            return connect_t, read_t

        raise AppriseInvalidData(
            "timeout must be None, a float, or a (connect, read) tuple"
        )

    @property
    def connected(self) -> bool:
        return self._sock is not None

    @property
    def is_tls(self) -> bool:
        return self._is_tls

    def close(self) -> None:
        """Close the socket and associated file wrappers."""
        try:
            if self._wfile is not None:
                with contextlib.suppress(Exception):
                    self._wfile.flush()
                with contextlib.suppress(Exception):
                    self._wfile.close()
        finally:
            self._wfile = None

        try:
            if self._rfile is not None:
                with contextlib.suppress(Exception):
                    self._rfile.close()
        finally:
            self._rfile = None

        if self._sock is not None:
            try:
                with contextlib.suppress(Exception):
                    self._sock.shutdown(socket.SHUT_RDWR)

                self._sock.close()
            finally:
                self._sock = None

        self._is_tls = False
        self._had_io = False
        self.local_addr = None
        self.remote_addr = None

    def _refresh_wrappers(self) -> None:
        """Rebuild file wrappers, required after TLS upgrade."""

        if self._sock is None:
            self._rfile = None
            self._wfile = None
            return

        self._rfile = self._sock.makefile("rb", buffering=0)
        self._wfile = self._sock.makefile("wb", buffering=0)

    def can_read(self, timeout: float = 0.0) -> Optional[bool]:
        """Return True if readable, False if not, None if closed or error."""
        if self._sock is None:
            return None
        try:
            r, _, x = select.select(
                [self._sock], [], [self._sock], float(timeout)
            )

        except OSError:
            self.close()
            return None

        if x:
            self.close()
            return None

        return bool(r)

    def can_write(self, timeout: float = 0.0) -> Optional[bool]:
        """Return True if writable, False if not, None if closed or error."""
        if self._sock is None:
            return None
        try:
            _, w, x = select.select(
                [], [self._sock], [self._sock], float(timeout)
            )
        except OSError:
            self.close()
            return None
        if x:
            self.close()
            return None
        return bool(w)

    def connect(self) -> None:
        """
        Establish TCP connection, optionally upgrade to TLS immediately if
        secure=True.
        """
        logger.trace(
            "Socket connect IN: host=%s port=%d secure=%s verify=%s",
            self.host,
            self.port,
            self.secure,
            self.verify,
        )
        self.close()

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            if self.bind_addr is not None or self.bind_port is not None:
                sock.bind(
                    (self.bind_addr or "127.0.0.1", int(self.bind_port or 0))
                )

            if self._connect_timeout is not None:
                sock.settimeout(self._connect_timeout)

            # Establish our connection
            sock.connect((self.host, self.port))

            # We control I/O blocking explicitly with select()
            sock.settimeout(None)

            self._sock = sock
            self._is_tls = False
            self._had_io = False

            if self.secure:
                self.start_tls()

            self.local_addr = self._sock.getsockname()
            self.remote_addr = self._sock.getpeername()
            self._refresh_wrappers()

            logger.debug(
                "Socket connected: local=%s remote=%s tls=%s",
                self.local_addr,
                self.remote_addr,
                self._is_tls,
            )

        except Exception as e:
            with contextlib.suppress(Exception):
                sock.close()

            self._sock = None
            self._had_io = False
            logger.debug("Socket connect exception: %s", e)
            raise AppriseSocketError(str(e)) from e

    def _server_hostname_for_tls(self) -> str:
        """
        Determine hostname used for SNI and hostname verification.

        If verify=True and host is an IP address, attempt reverse DNS lookup.
        """
        host = self.host

        if not self.verify:
            return host

        try:
            ipaddress.ip_address(host)
        except ValueError:
            return host

        try:
            name, _, _ = socket.gethostbyaddr(host)
            return name.rstrip(".") if name else host
        except Exception:
            return host

    def _build_ssl_context(self) -> ssl.SSLContext:
        """Build SSL context using certifi bundle when verify=True."""
        # Enforce TLS 1.2+ to avoid TLSv1/TLSv1.1 negotiation.
        # We explicitly enforce TLS >= 1.2 for Python 3.9+ compatibility.
        if self.verify:
            import certifi

            ctx = ssl.create_default_context(cafile=certifi.where())
            ctx.check_hostname = True
            ctx.verify_mode = ssl.CERT_REQUIRED
        else:
            # Still enforce modern TLS even when certificate verification is
            # disabled, since protocol downgrade is independent of trust.
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE

        # Enforce TLS 1.2+ to avoid TLSv1/TLSv1.1 negotiation.
        try:
            ctx.minimum_version = ssl.TLSVersion.TLSv1_2
        except Exception:
            # Fallback for very old Python/OpenSSL combinations
            with contextlib.suppress(Exception):
                ctx.options |= ssl.OP_NO_TLSv1
            with contextlib.suppress(Exception):
                ctx.options |= ssl.OP_NO_TLSv1_1

        # Disable TLS-level compression (mitigates CRIME-style attacks).
        with contextlib.suppress(Exception):
            ctx.options |= ssl.OP_NO_COMPRESSION

        return ctx

    def start_tls(self) -> None:
        """Upgrade an existing TCP connection to TLS."""
        if self._sock is None:
            raise AppriseSocketError("No active connection to upgrade")

        if self._is_tls:
            return

        server_hostname = self._server_hostname_for_tls()
        logger.trace("Starting TLS upgrade: sni=%s", server_hostname)

        try:
            ctx = self._build_ssl_context()
            tls_sock = ctx.wrap_socket(
                self._sock,
                server_hostname=server_hostname,
            )

            tls_sock.setblocking(False)
            self._sock = tls_sock
            self._is_tls = True

            self.local_addr = self._sock.getsockname()
            self.remote_addr = self._sock.getpeername()
            self._refresh_wrappers()

            logger.trace(
                "TLS upgrade complete: local=%s remote=%s",
                self.local_addr,
                self.remote_addr,
            )

        except ssl.SSLError as e:
            self.close()
            logger.debug("TLS negotiation exception: %s", e)
            raise AppriseSocketError(f"TLS negotiation failed: {e}") from e
        except OSError as e:
            self.close()
            logger.debug("TLS negotiation exception: %s", e)
            raise AppriseSocketError(str(e)) from e

    def _attempt_reconnect(
        self,
        retries: int,
        action: str,
        exc: Exception,
    ) -> bool:
        """
        Attempt to reconnect and allow the caller to retry.

        Args:
            retries: Remaining reconnect attempts permitted (<= 0 disables).
            action: A short label (e.g. "read" or "write") for logging.
            exc: The exception that triggered the reconnect attempt.

        Returns:
            True if a reconnect was performed and the caller should retry.
        """
        # Respect the caller's retry budget
        if int(retries) <= 0:
            return False

        # Only retry if we have previously completed useful I/O since the last
        # connect(). This prevents retrying the first failed read/write after
        # connect.
        if not self._had_io:
            return False

        logger.warning("Socket %s failed, reconnecting and retrying", action)
        logger.debug("Socket %s exception: %s", action, exc)

        try:
            self.close()
            self.connect()

        except Exception as e:
            logger.debug("Socket reconnect exception: %s", e)
            return False

        return True

    def read(
        self,
        max_bytes: int = 32768,
        blocking: bool = False,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
    ) -> bytes:
        """
        Read up to max_bytes bytes.

        blocking=False:
          - returns immediately with available data, or b"" if none

        blocking=True:
          - waits up to timeout seconds (or instance read timeout if timeout is
            None), then reads once
          - if both are None, waits indefinitely

        retries:
          - number of reconnect attempts permitted if the socket goes stale
            after prior successful I/O. Defaults to None (which takes value
            globally passed into the class)
        """
        if self._sock is None:
            return b""

        # Compute retry attempts; treat retries=0 as explicit 0
        retry_count = self.retries if retries is None else int(retries)
        attempts = max(0, retry_count) + 1

        # Derive wait timeout (None means wait indefinitely)
        wait_timeout = (
            self._read_timeout if timeout is None else float(timeout)
        )

        # We manage readiness via select, socket stays non-blocking
        self._sock.setblocking(False)

        while attempts:
            attempts -= 1

            try:
                if not blocking:
                    try:
                        data = self._sock.recv(int(max_bytes))
                        if data == b"":
                            raise AppriseSocketError(
                                "Connection lost during read"
                            )
                        self._had_io = True
                        return data
                    except (
                        BlockingIOError,
                        ssl.SSLWantReadError,
                        ssl.SSLWantWriteError,
                    ):
                        return b""

                # blocking=True path: wait for readability, then recv
                if wait_timeout is None:
                    # Wait indefinitely but periodically confirm socket health
                    while True:
                        ready = self.can_read(0.5)
                        if ready is None:
                            raise AppriseSocketError("Socket closed")

                        if ready:
                            break
                else:
                    ready = self.can_read(wait_timeout)
                    if not ready:
                        return b""

                # Even after select says readable, TLS may still raise
                # WANT_READ/WRITE. Loop until we either receive data, timeout,
                # or the socket closes.
                while True:
                    try:
                        data = self._sock.recv(int(max_bytes))
                        if data == b"":
                            raise AppriseSocketError(
                                "Connection lost during read"
                            )
                        self._had_io = True
                        return data

                    except (
                        ssl.SSLWantReadError,
                        ssl.SSLWantWriteError,
                        BlockingIOError,
                    ):
                        if wait_timeout is None:
                            continue

                        # Avoid busy loop
                        if not self.can_read(min(0.25, wait_timeout)):
                            return b""

            except (AppriseSocketError, OSError, ssl.SSLError) as e:
                # Normalise and log
                logger.warning("Socket read failed")
                logger.debug("Socket read exception: %s", e)

                # Only close on hard errors; WANT_READ/WRITE handled above
                if (
                    isinstance(e, OSError)
                    and not isinstance(e, ssl.SSLWantReadError)
                    and not isinstance(e, ssl.SSLWantWriteError)
                ):
                    self.close()

                err: Exception = e

                # Reconnect only if we've had prior useful I/O
                if self._attempt_reconnect(
                    retries=attempts,
                    action="read",
                    exc=err,
                ):
                    # In blocking mode with no timeout (wait indefinitely),
                    # perform an immediate read attempt after reconnect.
                    # This avoids relying solely on can_read(), and it keeps
                    # edge cases (like stale sockets) recoverable.
                    if (
                        blocking
                        and wait_timeout is None
                        and self._sock is not None
                    ):
                        try:
                            data = self._sock.recv(int(max_bytes))

                            if data == b"":
                                raise AppriseSocketError(
                                    "Connection lost during read"
                                )

                            self._had_io = True
                            return data

                        except (
                            BlockingIOError,
                            ssl.SSLWantReadError,
                            ssl.SSLWantWriteError,
                        ):
                            # No data yet; fall back to retry loop
                            pass

                    continue

                if isinstance(err, AppriseSocketError):
                    raise err from None
                raise AppriseSocketError(str(err)) from err

        raise AppriseSocketError("Socket read failed")

    def write(
        self,
        data: bytes,
        flush: bool = True,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
    ) -> int:
        """
        Write bytes to the socket.

        timeout:
          - if None, uses instance read timeout
          - if both are None, blocks until complete

        retries:
          - number of reconnect attempts permitted if the socket goes stale
            after prior successful I/O. Defaults to None (which takes value
            globally passed into the class)
        """
        if self._sock is None:
            raise AppriseSocketError("No active connection")

        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise AppriseInvalidData("write() expects bytes-like data")

        # Loop-based retry avoids recursion and keeps state obvious
        retry_count = self.retries if retries is None else int(retries)
        attempts = max(0, retry_count) + 1

        while attempts:
            attempts -= 1

            view = memoryview(data)
            total_sent = 0

            op_timeout = (
                self._read_timeout if timeout is None else float(timeout)
            )
            deadline = (
                None if op_timeout is None else (time.monotonic() + op_timeout)
            )

            try:
                self._sock.setblocking(deadline is None)

                while total_sent < len(view):
                    if deadline is not None:
                        remaining = deadline - time.monotonic()
                        if remaining <= 0:
                            raise AppriseSocketError("Timed out during write")
                        writable = self.can_write(remaining)
                        if not writable:
                            raise AppriseSocketError(
                                "Timed out waiting for writable socket"
                            )

                    sent = self._sock.send(view[total_sent:])
                    if sent <= 0:
                        raise AppriseSocketError(
                            "Connection lost during write"
                        )
                    total_sent += sent

                if flush and self._wfile is not None:
                    self._wfile.flush()

                if total_sent > 0:
                    self._had_io = True

                return total_sent

            except (AppriseSocketError, OSError) as e:
                logger.warning("Socket write failed")
                logger.debug("Socket write exception: %s", e)

                # Normalise: any OSError implies the socket is toast
                if isinstance(e, OSError):
                    self.close()

                if self._attempt_reconnect(
                    retries=attempts,
                    action="write",
                    exc=e,
                ):
                    continue

                if isinstance(e, AppriseSocketError):
                    raise
                raise AppriseSocketError(str(e)) from e

        raise AppriseSocketError("Socket write failed")



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime, timedelta

# Disable logging for a cleaner testing output
import logging
from timeit import default_timer

import pytest

from apprise import AppriseAsset, NotifyImageSize, NotifyType
from apprise.plugins import NotifyBase

logging.disable(logging.CRITICAL)


def test_notify_base():
    """
    API: NotifyBase() object

    """

    # invalid types throw exceptions
    with pytest.raises(TypeError):
        NotifyBase(**{"format": "invalid"})

    # invalid types throw exceptions
    with pytest.raises(TypeError):
        NotifyBase(**{"overflow": "invalid"})

    # Bad port information
    nb = NotifyBase(port="invalid")
    assert nb.port is None

    nb = NotifyBase(port=10)
    assert nb.port == 10

    assert isinstance(nb.url(), str)
    assert str(nb) == nb.url()

    with pytest.raises(NotImplementedError):
        # Each sub-module is that inherits this as a parent is required to
        # over-ride this function. So direct calls to this throws a not
        # implemented error intentionally
        nb.send("test message")

    # Throttle overrides..
    nb = NotifyBase()
    nb.request_rate_per_sec = 0.0
    start_time = default_timer()
    nb.throttle()
    elapsed = default_timer() - start_time
    # Should be a very fast response time since we set it to zero but we'll
    # check for less then 500 to be fair as some testing systems may be slower
    # then other
    assert elapsed < 0.5

    # Concurrent calls should achieve the same response
    start_time = default_timer()
    nb.throttle()
    elapsed = default_timer() - start_time
    assert elapsed < 0.5

    nb = NotifyBase()
    nb.request_rate_per_sec = 1.0

    # Set our time to now
    start_time = default_timer()
    nb.throttle()
    elapsed = default_timer() - start_time
    # A first call to throttle (Without telling it a time previously ran) does
    # not block for any length of time; it just merely sets us up for
    # concurrent calls to block
    assert elapsed < 0.5

    # Concurrent calls could take up to the rate_per_sec though...
    start_time = default_timer()
    nb.throttle(last_io=datetime.now())
    elapsed = default_timer() - start_time
    assert elapsed > 0.48 and elapsed < 1.5

    nb = NotifyBase()
    nb.request_rate_per_sec = 1.0

    # Set our time to now
    start_time = default_timer()
    nb.throttle(last_io=datetime.now())
    elapsed = default_timer() - start_time
    # because we told it that we had already done a previous action (now)
    # the throttle holds out until the right time has passed
    assert elapsed > 0.48 and elapsed < 1.5

    # Concurrent calls could take up to the rate_per_sec though...
    start_time = default_timer()
    nb.throttle(last_io=datetime.now())
    elapsed = default_timer() - start_time
    assert elapsed > 0.48 and elapsed < 1.5

    nb = NotifyBase()
    start_time = default_timer()
    nb.request_rate_per_sec = 1.0
    # Force a time in the past
    nb.throttle(last_io=(datetime.now() - timedelta(seconds=20)))
    elapsed = default_timer() - start_time
    # Should be a very fast response time since we set it to zero but we'll
    # check for less then 500 to be fair as some testing systems may be slower
    # then other
    assert elapsed < 0.5

    # Force a throttle time
    start_time = default_timer()
    nb.throttle(wait=0.5)
    elapsed = default_timer() - start_time
    assert elapsed > 0.48 and elapsed < 1.5

    # our NotifyBase wasn't initialized with an ImageSize so this will fail
    assert nb.image_url(notify_type=NotifyType.INFO) is None
    assert nb.image_path(notify_type=NotifyType.INFO) is None
    assert nb.image_raw(notify_type=NotifyType.INFO) is None

    # Color handling
    assert nb.color(notify_type="invalid") == AppriseAsset.default_html_color
    assert isinstance(
        nb.color(notify_type=NotifyType.INFO, color_type=None), str
    )
    assert isinstance(
        nb.color(notify_type=NotifyType.INFO, color_type=int), int
    )
    assert isinstance(
        nb.color(notify_type=NotifyType.INFO, color_type=tuple), tuple
    )

    # Ascii Handling
    assert nb.ascii(notify_type="invalid") == AppriseAsset.default_ascii_chars
    assert nb.ascii(NotifyType.INFO) == "[i]"
    assert nb.ascii(NotifyType.SUCCESS) == "[+]"
    assert nb.ascii(NotifyType.WARNING) == "[~]"
    assert nb.ascii(NotifyType.FAILURE) == "[!]"

    # Create an object
    nb = NotifyBase()
    # Force an image size since the default doesn't have one
    nb.image_size = NotifyImageSize.XY_256

    # We'll get an object this time around
    assert nb.image_url(notify_type=NotifyType.INFO) is not None
    assert nb.image_path(notify_type=NotifyType.INFO) is not None
    assert nb.image_raw(notify_type=NotifyType.INFO) is not None

    # Static function testing
    assert (
        NotifyBase.escape_html("<content>'\t \n</content>")
        == "&lt;content&gt;&apos;&emsp;&nbsp;\n&lt;/content&gt;"
    )

    assert (
        NotifyBase.escape_html(
            "<content>'\t \n</content>", convert_new_lines=True
        )
        == "&lt;content&gt;&apos;&emsp;&nbsp;<br/>&lt;/content&gt;"
    )

    # Test invalid data
    assert NotifyBase.split_path(None) == []
    assert NotifyBase.split_path(object()) == []
    assert NotifyBase.split_path(42) == []

    assert NotifyBase.split_path(
        "/path/?name=Dr%20Disrespect", unquote=False
    ) == [
        "path",
        "?name=Dr%20Disrespect",
    ]

    assert NotifyBase.split_path(
        "/path/?name=Dr%20Disrespect", unquote=True
    ) == [
        "path",
        "?name=Dr Disrespect",
    ]

    # a slash found inside the path, if escaped properly will not be broken
    # by split_path while additional concatinated slashes are ignored
    # FYI: %2F = /
    assert NotifyBase.split_path(
        "/%2F///%2F%2F////%2F%2F%2F////", unquote=True
    ) == [
        "/",
        "//",
        "///",
    ]

    # Test invalid data
    assert NotifyBase.parse_list(None) == []
    assert NotifyBase.parse_list(object()) == []
    assert NotifyBase.parse_list(42) == []

    result = NotifyBase.parse_list(
        ",path,?name=Dr%20Disrespect", unquote=False
    )
    assert isinstance(result, list)
    assert len(result) == 2
    assert "path" in result
    assert "?name=Dr%20Disrespect" in result

    result = NotifyBase.parse_list(",path,?name=Dr%20Disrespect", unquote=True)
    assert isinstance(result, list)
    assert len(result) == 2
    assert "path" in result
    assert "?name=Dr Disrespect" in result

    # by parse_list while additional concatinated slashes are ignored
    # FYI: %2F = /
    # In this lit there are actually 4 entries, however parse_list
    # eliminates duplicates in addition to unquoting content by default
    result = NotifyBase.parse_list(
        ",%2F,%2F%2F, , , ,%2F%2F%2F, %2F", unquote=True
    )
    assert isinstance(result, list)
    assert len(result) == 3
    assert "/" in result
    assert "//" in result
    assert "///" in result

    # Phone number parsing
    assert NotifyBase.parse_phone_no(None) == []
    assert NotifyBase.parse_phone_no(object()) == []
    assert NotifyBase.parse_phone_no(42) == []

    result = NotifyBase.parse_phone_no(
        "+1-800-123-1234,(800) 123-4567", unquote=False
    )
    assert isinstance(result, list)
    assert len(result) == 2
    assert "+1-800-123-1234" in result
    assert "(800) 123-4567" in result

    # %2B == +
    result = NotifyBase.parse_phone_no(
        "%2B1-800-123-1234,%2B1%20800%20123%204567", unquote=True
    )
    assert isinstance(result, list)
    assert len(result) == 2
    assert "+1-800-123-1234" in result
    assert "+1 800 123 4567" in result

    # Give nothing, get nothing
    assert NotifyBase.escape_html("") == ""
    assert NotifyBase.escape_html(None) == ""
    assert NotifyBase.escape_html(object()) == ""

    # Test quote
    assert NotifyBase.unquote("%20") == " "
    assert NotifyBase.quote(" ") == "%20"
    assert NotifyBase.unquote(None) == ""
    assert NotifyBase.quote(None) == ""


def test_notify_base_urls():
    """
    API: NotifyBase() URLs

    """

    # Test verify switch whih is used as part of the SSL Verification
    # by default all SSL sites are verified unless this flag is set to
    # something like 'No', 'False', 'Disabled', etc.  Boolean values are
    # pretty forgiving.
    results = NotifyBase.parse_url("https://localhost:8080/?verify=No")
    assert "verify" in results
    assert results["verify"] is False

    results = NotifyBase.parse_url("https://localhost:8080/?verify=Yes")
    assert "verify" in results
    assert results["verify"] is True

    # The default is to verify
    results = NotifyBase.parse_url("https://localhost:8080")
    assert "verify" in results
    assert results["verify"] is True

    # Password Handling

    # pass keyword over-rides default password
    results = NotifyBase.parse_url("https://user:pass@localhost")
    assert "password" in results
    assert results["password"] == "pass"

    # pass keyword over-rides default password
    results = NotifyBase.parse_url(
        "https://user:pass@localhost?pass=newpassword"
    )
    assert "password" in results
    assert results["password"] == "newpassword"

    # password keyword can also optionally be used
    results = NotifyBase.parse_url(
        "https://user:pass@localhost?password=passwd"
    )
    assert "password" in results
    assert results["password"] == "passwd"

    # pass= override password=
    # password keyword can also optionally be used
    results = NotifyBase.parse_url(
        "https://user:pass@localhost?pass=pw1&password=pw2"
    )
    assert "password" in results
    assert results["password"] == "pw1"

    # Options
    results = NotifyBase.parse_url("https://localhost?format=invalid")
    assert "format" not in results
    results = NotifyBase.parse_url("https://localhost?format=text")
    assert "format" in results
    assert results["format"] == "text"
    results = NotifyBase.parse_url("https://localhost?format=markdown")
    assert "format" in results
    assert results["format"] == "markdown"
    results = NotifyBase.parse_url("https://localhost?format=html")
    assert "format" in results
    assert results["format"] == "html"

    results = NotifyBase.parse_url("https://localhost?overflow=invalid")
    assert "overflow" not in results
    results = NotifyBase.parse_url("https://localhost?overflow=upstream")
    assert "overflow" in results
    assert results["overflow"] == "upstream"
    results = NotifyBase.parse_url("https://localhost?overflow=split")
    assert "overflow" in results
    assert results["overflow"] == "split"
    results = NotifyBase.parse_url("https://localhost?overflow=truncate")
    assert "overflow" in results
    assert results["overflow"] == "truncate"

    # User Handling

    # user keyword over-rides default password
    results = NotifyBase.parse_url("https://user:pass@localhost")
    assert "user" in results
    assert results["user"] == "user"

    # user keyword over-rides default password
    results = NotifyBase.parse_url("https://user:pass@localhost?user=newuser")
    assert "user" in results
    assert results["user"] == "newuser"

    # Test invalid urls
    assert NotifyBase.parse_url("https://:@/") is None
    assert NotifyBase.parse_url("http://:@") is None
    assert NotifyBase.parse_url("http://@") is None
    assert NotifyBase.parse_url("http:///") is None
    assert NotifyBase.parse_url("http://:test/") is None
    assert NotifyBase.parse_url("http://pass:test/") is None


def test_notify_base_runtime_deps():
    """NotifyBase.runtime_deps() returns an empty tuple by default."""

    assert NotifyBase.runtime_deps() == ()
    assert isinstance(NotifyBase.runtime_deps(), tuple)


def test_notify_base_enable_disable():
    """NotifyBase.enable() and disable() toggle the enabled flag."""

    class ToggleNotification(NotifyBase):
        protocol = "toggle"

        def notify(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return "toggle://"

    # Starts enabled
    assert ToggleNotification.enabled is True

    # disable() sets it False
    ToggleNotification.disable()
    assert ToggleNotification.enabled is False

    # enable() restores it
    ToggleNotification.enable()
    assert ToggleNotification.enabled is True

    # Calling disable() twice is idempotent
    ToggleNotification.disable()
    ToggleNotification.disable()
    assert ToggleNotification.enabled is False

    # Calling enable() twice is idempotent
    ToggleNotification.enable()
    ToggleNotification.enable()
    assert ToggleNotification.enabled is True

    # Changes are on the class, not on instances
    instance = ToggleNotification(host="localhost")
    ToggleNotification.disable()
    assert instance.enabled is False
    ToggleNotification.enable()
    assert instance.enabled is True


def test_notify_base_runtime_deps_override():
    """Subclasses can override runtime_deps() to declare their dependencies."""

    class DepFreePlugin(NotifyBase):
        protocol = "depfree"

        def notify(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return "depfree://"

    class SingleDepPlugin(NotifyBase):
        protocol = "singledep"

        def notify(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return "singledep://"

        @staticmethod
        def runtime_deps():
            return ("fakelibA",)

    class MultiDepPlugin(NotifyBase):
        protocol = "multidep"

        def notify(self, *args, **kwargs):
            return True

        def url(self, **kwargs):
            return "multidep://"

        @staticmethod
        def runtime_deps():
            return ("fakelibA", "fakelibB")

    # No override -> inherits the empty tuple from NotifyBase
    assert DepFreePlugin.runtime_deps() == ()

    # Single-dep override
    assert SingleDepPlugin.runtime_deps() == ("fakelibA",)

    # Multi-dep override preserves order
    deps = MultiDepPlugin.runtime_deps()
    assert deps == ("fakelibA", "fakelibB")
    assert deps[0] == "fakelibA"
    assert deps[1] == "fakelibB"



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise.plugins.pagertree import NotifyPagerTree

logging.disable(logging.CRITICAL)

# a test UUID we can use
INTEGRATION_ID = "int_xxxxxxxxxxx"

# Our Testing URLs
apprise_url_tests = (
    (
        "pagertree://",
        {
            # Missing Integration ID
            "instance": TypeError,
        },
    ),
    # Invalid Integration ID
    (
        "pagertree://%s" % ("+" * 24),
        {
            "instance": TypeError,
        },
    ),
    # Minimum requirements met
    (
        f"pagertree://{INTEGRATION_ID}",
        {
            "instance": NotifyPagerTree,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "pagertree://i...x?",
        },
    ),
    # change the integration id
    (
        f"pagertree://{INTEGRATION_ID}?integration=int_yyyyyyyyyy",
        {
            "instance": NotifyPagerTree,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "pagertree://i...y?",
        },
    ),
    # entries specified on the URL will over-ride the host (integration id)
    (
        f"pagertree://{INTEGRATION_ID}?id=int_zzzzzzzzzz",
        {
            "instance": NotifyPagerTree,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "pagertree://i...z?",
        },
    ),
    # Integration ID + bad url
    (
        "pagertree://:@/",
        {
            "instance": TypeError,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}",
        {
            "instance": NotifyPagerTree,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}",
        {
            "instance": NotifyPagerTree,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}",
        {
            "instance": NotifyPagerTree,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}?urgency=low",
        {
            # urgency override
            "instance": NotifyPagerTree,
        },
    ),
    (
        f"pagertree://?id={INTEGRATION_ID}&urgency=low",
        {
            # urgency override and id= (for integration)
            "instance": NotifyPagerTree,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}?tags=production,web",
        {
            # tags
            "instance": NotifyPagerTree,
        },
    ),
    (
        f"pagertree://{INTEGRATION_ID}?action=resolve&thirdparty=123",
        {
            # test resolve
            "instance": NotifyPagerTree,
        },
    ),
    # Custom values
    (
        f"pagertree://{INTEGRATION_ID}?+pagertree-token=123&:env=prod&-m=v",
        {
            # minimum requirements and support custom key/value pairs
            "instance": NotifyPagerTree,
        },
    ),
)


def test_plugin_pagertree_urls():
    """NotifyPagerTree() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_pagertree_general(mock_post):
    """NotifyPagerTree() General Checks."""

    # Prepare Mock
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok

    # Invalid thirdparty id
    with pytest.raises(TypeError):
        NotifyPagerTree(integration=INTEGRATION_ID, thirdparty="   ")
